# -*- coding: utf-8 -*-
"""Abstract base classes and utilities for moments"""

import abc
import collections
from typing import Tuple, List, Dict, Iterator, Union

import numpy as np

from . import functions

__all__ = 'Moments', 'Index', 'ReconstructionQuery'

Index = Tuple[int, int, int]  # A three dimensional moment index


class ReconstructionQuery:

    def __init__(self, max_order: int, points: np.ndarray):
        self._max_order = max_order
        self._points = points

    @property
    def max_order(self) -> int:
        return self._max_order

    @property
    def points(self):
        return self._points


class Moments(functions.State, metaclass=abc.ABCMeta):
    """A class representing three dimensional moments"""

    # pylint: disable=invalid-name

    @classmethod
    def from_indexed(cls, indexed, max_order: int, dtype=float) -> 'Moments':
        moments = cls(max_order, dtype=dtype)
        for idx in moments.iter_indices():
            moments[idx] = indexed.__getitem__(idx)
        return moments

    @property
    @abc.abstractmethod
    def dtype(self):
        """Get the number type of the moments (typically float or complex)"""

    @abc.abstractmethod
    def __getitem__(self, index: Index):
        """Get the moment of the given index"""

    @abc.abstractmethod
    def __setitem__(self, index: Union[int, Tuple], value):
        """Set the moment of the given index"""

    def to_matrix(self) -> np.array:  # pylint: disable=no-self-use
        """Return the moments in a matrix.  Not all moment classes support this as they may have
        special indexing schemes in which case an AttributeError will be raised"""
        raise AttributeError('Does not support conversion to matrix')

    @property
    @abc.abstractmethod
    def max_order(self):
        """Get the maximum order of the moments"""

    @abc.abstractmethod
    def moment(self, n: int, l: int, m: int):
        """Get the n, l, m^th moment"""

    @abc.abstractmethod
    def linear_index(self, index: Index) -> int:
        """Get a linearised index from the passed triple index"""

    @abc.abstractmethod
    def iter_indices(self) -> Iterator[Index]:
        """Iterate through the valid indices of these moments"""

    @abc.abstractmethod
    def value_at(self, x: np.array, max_order: int = None):
        """Reconstruct the value at x from the moments

        :param x: the point to get the value at
        :param max_order: the maximum order to go up to (defaults to the maximum order of these
            moments)
        """

    @abc.abstractmethod
    def get_mask(self) -> 'Moments':
        """Get an empty set of moments all set to None that can be used as a mask to fix values e.g. for optimisation"""

    def grid_values(self, num_samples: int, zero_outside_domain=True):
        """Get a grid and corresponding values of the moments reconstructed at the gridpoints"""
        # Create a coordinate grid
        spacing = np.linspace(-1., 1., num_samples)
        grid = np.array(np.meshgrid(spacing, spacing, spacing))
        grid_points = grid.reshape(3, -1).T

        if zero_outside_domain:
            # Calculate the lengths squared and get the corresponding indexes
            length_sq = (grid_points**2).sum(axis=1)
            valid_idxs = np.argwhere(length_sq < 1)

            # Now calculate the grid values at those points, the rest are 0
            grid_vals = np.zeros(grid_points.shape[0])
            values = self.value_at(grid_points[valid_idxs][:, 0, :])
            np.put(grid_vals, valid_idxs, values, mode='raise')
        else:
            # Do all points, even those outside the domain
            grid_vals = self.value_at(grid_points)

        # Reshape into n * n * n array
        return grid, grid_vals.reshape((grid.shape[1:]))

    def reconstruct(
        self,
        query: ReconstructionQuery,
        order=None,  # pylint: disable=unused-argument
        zero_outside_domain=True  # pylint: disable=unused-argument
    ):
        return self.value_at(query.points)

    @classmethod
    def get_grid(cls, num_samples: int, restrict_to_domain=True) -> np.ndarray:
        # Create a coordinate grid
        spacing = np.linspace(-1., 1., num_samples)
        grid = np.array(np.meshgrid(spacing, spacing, spacing))
        grid_points = grid.reshape(3, -1).T

        if restrict_to_domain:
            # Calculate the lengths squared and get the corresponding indexes
            length_sq = (grid_points**2).sum(axis=1)
            valid_idxs = np.argwhere(length_sq < 1)
            grid_points = grid_points[valid_idxs][:, 0, :]

        return grid_points

    @classmethod
    def create_reconstruction_query_from_grid(
        cls, order: int, num_sample: int, restrict_to_domain=True
    ) -> ReconstructionQuery:
        grid = cls.get_grid(num_sample, restrict_to_domain=restrict_to_domain)
        return cls.create_reconstruction_query(grid, order)

    @classmethod
    def create_reconstruction_query(cls, points: np.ndarray, order: int) -> ReconstructionQuery:
        return ReconstructionQuery(order, points)


ProductTerm = collections.namedtuple('ProductTerm', 'index terms')
Product = collections.namedtuple('Product', 'coeff terms')


class MomentsCalculator(functions.Function, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def create_random(self, max_order: int = None):
        """Create a random set of moments (optionally) up to a certain order."""


class MomentsPolynomial:
    """Represents a polyonimal of moments.  The terms are products of moments with a prefactor"""
    __slots__ = ('_terms',)

    def __init__(self):
        self._terms: List[Product] = []

    def __str__(self):
        sum_parts = []
        for prefactor, product in self._terms:
            powers = self.collect_powers(product)
            product_parts = [str(prefactor)]
            product_parts.extend(
                'm{},{},{}^{}'.format(indices[0], indices[1], indices[2], power) for indices, power in powers.items()
            )

            string = ' '.join(product_parts)
            sum_parts.append(string)

        return ' + '.join(sum_parts)

    def append(self, term: Product):
        """Append a term to the polynomial"""
        self._terms.append(term)

    def evaluate(self, moments):
        """Evaluate the polynomial for a given set of moments"""
        total = 0.

        for term in self._terms:
            partial = term.coeff
            for product_term in term.terms:
                partial *= moments[product_term]
            total += partial

        return total

    @staticmethod
    def collect_powers(product: List[Tuple]) -> Dict[Tuple, int]:
        powers = collections.defaultdict(int)
        for indices in product:
            powers[indices] += 1
        return powers


class ProductBuilder:
    """Helper to build a product of moments that can form part of a moments polynomial"""

    def __init__(self, coeff):
        self._coeff = coeff
        self._terms = []

    def add(self, index):
        """Add a moment to the product using its index"""
        self._terms.append(index)

    def build(self) -> Product:
        """Build the product into a fixed tuple"""
        return Product(self._coeff, tuple(self._terms))



# -*- coding: utf-8 -*-
import numpy
from scipy.spatial.transform import Rotation


def center(points: numpy.array) -> numpy.array:
    """Centre the given points by shifting the centre of mass to zero"""
    return points - points.sum(axis=0) / len(points)


def randomly_rotate(points: numpy.array) -> numpy.array:
    """Randomly rotate a set of points"""
    rot = Rotation.random()
    return rot.apply(points)



# -*- coding: utf-8 -*-
"""Module for calculating geometric moments"""
import collections
import functools
import numbers
from typing import Union, Tuple, Optional, Iterator

import numpy as np
from numpy import ma
import sympy

from . import base_moments
from . import functions
from . import utils

__all__ = 'GeometricMoments', 'GeometricMomentsCalculator'

# pylint: disable=invalid-name

# Some constants to make understanding indexing easier
X = 0
Y = 1
Z = 2

GeometricIndex = collections.namedtuple('GeometricIndex', 'p q r')


class GeometricMoments(base_moments.Moments):

    @staticmethod
    def num_moments(max_order: int) -> int:
        """Get the total number of moments up to the given maximum order"""
        return len(tuple(iter_indices(max_order)))

    @classmethod
    def from_vector(cls, vec: np.array, max_order: int) -> 'GeometricMoments':
        moms_mtx = np.zeros((max_order + 1, max_order + 1, max_order + 1))
        indices = np.array(list(iter_indices(max_order))).T
        linear = np.ravel_multi_index(indices, moms_mtx.shape)

        # Put the vector values into the moments matrix
        np.put(moms_mtx, linear, vec)

        return GeometricMoments(moms_mtx)

    def __init__(self, moments: np.array):
        # Just trying this for now, let's see if it's possible to not force self._moments to be a numpy array
        max_order = moments.shape[0] - 1
        # Mask everything and then unmask the ones we want
        mask = np.ones(moments.shape, dtype=np.int)
        idx_array = np.empty(moments.shape, dtype=np.int)
        for i, indices in enumerate(iter_indices(max_order)):
            mask[indices] = 0
            idx_array[indices] = i

        self._moments = ma.array(moments, mask=mask)
        self._idx_array = ma.array(idx_array, mask=mask)

    def __getitem__(self, index: base_moments.Index):
        """Get the moment of the given index"""
        return self._moments[index]

    def __setitem__(self, index: base_moments.Index, value):
        """Get the moment of the given index"""
        self._moments[index] = value

    @property
    def dtype(self) -> type:
        try:
            return self._moments.dtype
        except AttributeError:
            return type(self._moments[0, 0, 0])

    @property
    def max_order(self) -> int:
        return self._moments.shape[0] - 1

    @property
    def vector(self) -> np.array:
        return self._moments.compressed()

    @property
    def moments(self) -> np.array:
        return self._moments

    def to_matrix(self) -> np.array:
        return self._moments

    def moment(self, p: int, q: int, r: int) -> float:
        return self._moments[p, q, r]

    def linear_index(self, index: base_moments.Index) -> int:
        return self._idx_array[index]

    def iter_indices(self) -> Iterator[base_moments.Index]:
        """Iterate through the valid indices of these moments"""
        yield from iter_indices(self.max_order)

    def value_at(self, x: np.array, max_order: int = None):
        """Reconstruct the value at x from the moments

        :param x: the point to get the value at
        :param max_order: the maximum order to go up to (defaults to the maximum order of these
            moments)
        """
        shape = self._moments.shape
        value = 0.
        for p in range(shape[0]):
            for q in range(shape[1]):
                for r in range(shape[2]):
                    value += self._moments[p, q, r] * (x**(p, q, r)).prod(axis=-1)

        return value

    def get_mask(self) -> 'GeometricMoments':
        geom_moms = np.empty(self._moments.shape, dtype=object)
        geom_moms.fill(None)
        return GeometricMoments(geom_moms)

    def get_builder(self, mask=None) -> Optional[functions.Function]:
        return functions.FromVectorBuilder(GeometricMoments, kwargs=dict(max_order=self.max_order))


def linear_index(max_order: int, index: base_moments.Index) -> int:
    if np.sum(index) > max_order:
        raise ValueError(f'The index passed ({index}) is of higher than the max order ({max_order})')

    for i, indices in enumerate(iter_indices(max_order)):
        if indices == index:
            return i


def iter_indices(max_order: int) -> Iterator[base_moments.Index]:
    for i in utils.inclusive(max_order):
        for j in utils.inclusive(max_order - i):
            for k in utils.inclusive(max_order - i - j):
                yield GeometricIndex(i, j, k)


class GeometricMomentsCalculator(base_moments.MomentsCalculator):
    """Function that calculates geometric moments up to some maximum order"""

    supports_jacobian = False

    def __init__(self, max_order: int):
        super().__init__()
        self._max_order = max_order

    def output_length(self, _in_state: functions.State) -> int:
        return GeometricMoments.num_moments(self._max_order)

    def evaluate(self, features: functions.State, *, get_jacobian=False) -> \
            Union[GeometricMoments, Tuple[GeometricMoments, np.ndarray]]:
        result = geometric_moments(features, self._max_order, get_jacobian, dtype=features.vector.dtype)

        if get_jacobian:
            return GeometricMoments(result[0]), result[1]

        return GeometricMoments(result)

    def create_random(self, max_order: int = None):
        max_val = 5.
        max_order = max_order or self._max_order
        return GeometricMoments(max_val * (2 * np.random.rand(max_order, max_order, max_order) - 1))


@functools.singledispatch
def geometric_moments(state: functions.State, max_order: int, get_jacobian: bool, dtype) -> np.ndarray:
    raise TypeError("Don't know how to calculate geometric moments for type '{}'".format(type(state).__name__))


@geometric_moments.register(functions.WeightedDelta)
def _(delta: functions.WeightedDelta, max_order: int, get_jacobian: bool, dtype):
    order = max_order

    moms = np.empty((order + 1, 3), dtype=dtype)
    moms[0] = 1  # 0^th power always 1.
    moms[1] = delta.pos  # 1^st power always the position itself

    # Calculate each x, y, z raise to powers up to the upper bound
    for power in utils.inclusive(2, order, 1):
        moms[power] = np.multiply(moms[power - 1, :], delta.pos)

    moments = delta.weight * utils.outer_product(moms[:, 0], moms[:, 1], moms[:, 2])

    if get_jacobian:
        indices = tuple(iter_indices(max_order))
        jacobian = np.zeros((len(indices), delta.LENGTH), dtype=dtype)

        for i, (p, q, r) in enumerate(indices):
            jacobian[i, delta.X] = p * moments[p - 1, q, r]
            jacobian[i, delta.Y] = q * moments[p, q - 1, r]
            jacobian[i, delta.Z] = r * moments[p, q, r - 1]
            jacobian[i, delta.WEIGHT] = moments[p, q, r] / delta.weight

        return moments, jacobian

    return moments


@geometric_moments.register(functions.WeightedGaussian)
def _(gaussian: functions.WeightedGaussian, max_order: int, get_jacobian: bool, dtype):
    """Calculate the geometric moments for a 3D Gaussian"""
    O = max_order

    partial_moments = np.empty((O + 1, 3), dtype=dtype)
    partial_moments[0, :] = 1.0  # 0^th order, mass is multiplied in at end

    for order in utils.inclusive(0, O):
        partial_moments[order, :] = np.array(gaussian_moment[order](gaussian.pos, gaussian.sigma))

    moments = utils.outer_product(partial_moments[:, 0], partial_moments[:, 1], partial_moments[:, 2])

    if get_jacobian:
        dg = np.empty((
            O + 1,  # Order
            2,  # d/dmu, d/dsigma
            3,  # x, y, z
        ))

        for order in utils.inclusive(max_order):
            dg[order, :, :] = np.array(gaussian_moment_derivatives[order](gaussian.pos, gaussian.sigma))

        # Let's get the indices and create a Jacobian matrix
        indices = tuple(iter_indices(max_order))
        jacobian = np.zeros((len(indices), gaussian.LENGTH), dtype=dtype)

        for i, (p, q, r) in enumerate(indices):
            # Positions
            jacobian[i, gaussian.X] = gaussian.weight * dg[p, 0, 0] * partial_moments[q, 1] * partial_moments[r, 2]
            jacobian[i, gaussian.Y] = gaussian.weight * dg[q, 0, 1] * partial_moments[p, 0] * partial_moments[r, 2]
            jacobian[i, gaussian.Z] = gaussian.weight * dg[r, 0, 2] * partial_moments[p, 0] * partial_moments[q, 1]

            # Sigma
            jacobian[i, gaussian.SIGMA] = gaussian.weight * (
                    dg[p, 1, 0] * partial_moments[q, 1] * partial_moments[r, 2] + \
                    partial_moments[p, 0] * dg[q, 1, 1] * partial_moments[r, 2] + \
                    partial_moments[p, 0] * partial_moments[q, 1] * dg[r, 1, 2]
            )

        # Weights
        jacobian[:, gaussian.WEIGHT] = moments[:, :, :]

        return gaussian.weight * moments, jacobian

    return gaussian.weight * moments


@geometric_moments.register(functions.Features)
def _(environment: functions.Features, max_order: int, get_jacobian, dtype):
    """Calculate the geometric moments for a set of features"""
    O = max_order
    idx = 0

    moments = np.empty((O + 1, O + 1, O + 1), dtype=dtype)
    moments.fill(0.)

    jacobians = []
    for feature in environment.features:
        if get_jacobian:
            moms, jac = geometric_moments(feature, max_order, get_jacobian, dtype)
            jacobians.append(jac)
        else:
            moms = geometric_moments(feature, max_order, get_jacobian, dtype)

        moments += moms

        idx += len(feature)

    if get_jacobian:
        return moments, np.concatenate(jacobians, axis=1)

    return moments


def from_gaussians(
    max_order: int,
    positions: np.ndarray,
    sigmas: Union[numbers.Number, np.array] = 0.4,
    weights: Union[numbers.Number, np.array] = 1.
) -> GeometricMoments:
    """Calculate the geometric moments for a collection of Gaussians at the given positions with
    the passed parameters.

    :param positions: the positions of the Gaussians
    :param sigmas: the standard deviations
    :param max_order: the maximum order to calculate moments up to
    :param weights: the masses of the Gaussians (or probabilities)
    :return: a max_order * max_order * max_order array of moments
    """
    positions = np.array(positions)
    shape = positions.shape[0]
    sigmas = _to_array(sigmas, shape)
    weights = _to_array(weights, shape)

    features = functions.Features()
    for pos, sigma, weight in zip(positions, sigmas, weights):
        features.add(functions.WeightedGaussian(pos, sigma, weight))

    return GeometricMomentsCalculator(max_order)(features)


def from_deltas(
    max_order: int,
    positions: np.array,
    weights: Union[numbers.Number, np.array] = 1,
    get_jacobian=False,
) -> Union[GeometricMoments, np.array]:
    """
    Calculate the geometric moments for a collection of delta functions at the given positions with
    the given weights.

    :param max_order: the maximum order to calculate moments up to
    :param positions: the positions of the delta functions
    :param weights: the weights of the delta functions
    :param get_jacobian: return the Jacobian along with the geometric moments
    :return: a max_order * max_order * max_order array of moments
    """
    if not isinstance(positions, np.ndarray):
        positions = np.array(positions)

    if not len(positions.shape) == 2 or positions.shape[1] != 3:
        raise ValueError(
            "Positions must be a list of vectors or a numpy array with shape [n, 3], not: '{}".format(positions.shape)
        )

    shape = positions.shape[0]
    weights = _to_array(weights, shape)

    features = functions.Features()
    for pos, weight in zip(positions, weights):
        features.add(functions.WeightedDelta(pos, weight))

    return GeometricMomentsCalculator(max_order)(features, jacobian=get_jacobian)


def from_deltas_analytic(max_order: int, num_particles: int, pos_symbols=None, weight_symbols=None):
    r = pos_symbols or sympy.IndexedBase('r')  # Positions of particles pylint: disable=invalid-name
    w = weight_symbols or sympy.IndexedBase('w')  # Weights of particles pylint: disable=invalid-name
    ubound = max_order + 1  # Upper bound

    # pylint: disable=invalid-name

    moments = sympy.MutableDenseNDimArray.zeros(ubound, ubound, ubound)

    for idx in range(num_particles):
        moms = sympy.MutableDenseNDimArray.zeros(ubound, 3)
        moms[0, :] = sympy.Array([1, 1, 1])  # 0^th power always 1.

        # Calculate each x, y, z raise to powers up to the upper bound
        for power in range(1, ubound):
            moms[power, 0] = moms[power - 1, 0] * r[idx, 0]
            moms[power, 1] = moms[power - 1, 1] * r[idx, 1]
            moms[power, 2] = moms[power - 1, 2] * r[idx, 2]

        moments += w[idx] * sympy.tensorproduct(moms[:, 0], moms[:, 1], moms[:, 2])

    return moments


def gaussian_moment_0(mu: float, sigma: float = 1.) -> float:  # pylint: disable=unused-argument
    return 1.


def gaussian_moment_0_derivative(mu: float, sigma: float = 1.) -> Tuple[float, float]:  # pylint: disable=unused-argument
    return 0., 0.


def gaussian_moment_1(mu: float, sigma: float = 1.) -> float:  # pylint: disable=unused-argument
    return mu


def gaussian_moment_1_derivative(mu: float, sigma: float = 1.) -> Tuple[float, float]:  # pylint: disable=unused-argument
    return 1., 0.


def gaussian_moment_2(mu: float, sigma: float = 1.) -> float:
    return mu**2 + sigma**2


def gaussian_moment_2_derivative(mu: float, sigma: float = 1.) -> Tuple[float, float]:
    return 2 * mu, 2 * sigma


def gaussian_moment_3(mu: float, sigma: float = 1.) -> float:
    return mu**3 + 3 * mu * sigma**2


def gaussian_moment_3_derivative(mu: float, sigma: float = 1.) -> Tuple[float, float]:
    return 3 * mu**2 + 3 * sigma**2, 6 * mu * sigma


def gaussian_moment_4(mu: float, sigma: float = 1.) -> Union[numbers.Number, np.ndarray]:
    return mu ** 4 + \
           6 * mu ** 2 * sigma ** 2 + \
           3 * sigma ** 4


def gaussian_moment_4_derivative(mu: float, sigma: float = 1.) -> Tuple[float, float]:
    dmu = 4 * mu ** 3 + \
          12 * mu * sigma ** 2
    dsigma = 12 * mu ** 2 * sigma + \
             12 * sigma ** 3

    return dmu, dsigma


def gaussian_moment_5(mu: float, sigma: float = 1.):
    return mu ** 5 + \
           10 * mu ** 3 * sigma ** 2 + \
           5 * mu * 3 * sigma ** 4


def gaussian_moment_5_derivative(mu: float, sigma: float = 1.) -> Tuple[float, float]:
    dmu = 5 * mu ** 4 + \
          30 * mu ** 2 * sigma ** 2 + \
          15 * sigma ** 4

    dsigma = 20 * mu ** 3 * sigma + \
             60 * mu * sigma ** 3

    return dmu, dsigma


def gaussian_moment_6(mu: float, sigma: float = 1.) -> float:
    mom = mu ** 6 + \
          15 * mu ** 4 * sigma ** 2 + \
          15 * mu ** 2 * 3 * sigma ** 4 + \
          15 * sigma ** 6

    return mom


def gaussian_moment_6_derivative(mu: float, sigma: float = 1.) -> Tuple[float, float]:
    dmu = 6 * mu ** 5 + \
          60 * mu ** 3 * sigma ** 2 + \
          90 * mu * sigma ** 4
    dsigma = 30 * mu ** 4 * sigma + \
             180 * mu ** 2 * sigma ** 3 + \
             90 * sigma ** 5

    return dmu, dsigma


def gaussian_moment_7(mu: float, sigma: float = 1.) -> float:
    return mu ** 7 + \
           21 * mu ** 5 * sigma ** 2 + \
           35 * mu ** 3 * 3 * sigma ** 4 + \
           7 * mu * 15 * sigma ** 6


def gaussian_moment_7_derivative(mu: float, sigma: float = 1.) -> Tuple[float, float]:
    dmu = 7 * mu ** 6 + \
          105 * mu ** 4 * sigma ** 2 + \
          315 * mu ** 2 * sigma ** 4 + \
          105 * sigma ** 6
    dsigma = 42 * mu ** 5 * sigma + \
             420 * mu ** 3 * sigma ** 3 + \
             630 * mu * sigma ** 5

    return dmu, dsigma


def gaussian_moment_8(mu: float, sigma: float = 1.) -> float:
    return mu ** 8 + \
           28 * mu ** 6 * sigma ** 2 + \
           70 * mu ** 4 * 3 * sigma ** 4 + \
           28 * mu ** 2 * 15 * sigma ** 6 + \
           105 * sigma ** 8


def gaussian_moment_8_derivative(mu: float, sigma: float = 1.) -> Tuple[float, float]:
    dmu = 8 * mu ** 7 + \
          168 * mu ** 5 * sigma ** 2 + \
          840 * mu ** 3 * sigma ** 4 + \
          840 * mu * sigma ** 6

    dsigma = 56 * mu ** 6 * sigma + \
             840 * mu ** 4 * sigma ** 3 + \
             2520 * mu ** 2 * sigma ** 5 + \
             840 * sigma ** 7

    return dmu, dsigma


def gaussian_moment_9(mu: float, sigma: float = 1.) -> float:
    return mu ** 9 + \
           36 * mu ** 7 * sigma ** 2 + \
           126 * mu ** 5 * 3 * sigma ** 4 + \
           84 * mu ** 3 * 15 * sigma ** 6 + \
           9 * mu * 105 * sigma ** 8


def gaussian_moment_10(mu: float, sigma: float = 1.) -> float:
    return mu ** 10 + \
           45 * mu ** 8 * sigma ** 2 + \
           210 * mu ** 6 * 3 * sigma ** 4 + \
           210 * mu ** 4 * 15 * sigma ** 6 + \
           45 * mu ** 2 * 105 * mu * sigma ** 8 + \
           945 * sigma ** 10


def gaussian_moment_11(mu: float, sigma: float = 1.) -> float:
    return mu ** 11 + \
           55 * mu ** 9 * sigma ** 2 + \
           330 * mu ** 7 * 3 * sigma ** 4 + \
           462 * mu ** 5 * 15 * sigma ** 6 + \
           165 * mu ** 3 * 105 * sigma ** 8 + \
           11 * mu * sigma ** 10


def gaussian_moment_12(mu: float, sigma: float = 1.) -> float:
    return mu ** 12 + \
           66 * mu ** 10 * sigma ** 2 + \
           495 * mu ** 8 * 3 * sigma ** 4 + \
           924 * mu ** 6 * 15 * sigma ** 6 + \
           495 * mu ** 4 * 105 * sigma ** 8 + \
           66 * mu ** 2 * 945 * sigma ** 10 + \
           10395 * sigma ** 12


def gaussian_moment_13(mu: float, sigma: float = 1.) -> float:
    return mu ** 13 + \
           78 * mu ** 11 * sigma ** 2 + \
           715 * mu ** 9 * 3 * sigma ** 4 + \
           1716 * mu ** 7 * 15 * sigma ** 6 + \
           1287 * mu ** 5 * 105 * sigma ** 8 + \
           286 * mu ** 3 * 945 * sigma ** 10 + \
           13 * mu * 10395 * sigma ** 12


def gaussian_moment_14(mu: float, sigma: float = 1.) -> float:
    return mu ** 14 + \
           91 * mu ** 12 * sigma ** 2 + \
           1001 * mu ** 10 * 3 * sigma ** 4 + \
           3003 * mu ** 8 * 15 * sigma ** 6 + \
           3003 * mu ** 6 * 105 * sigma ** 8 + \
           1001 * mu ** 4 * 945 * sigma ** 10 + \
           91 * mu ** 2 * 10395 * sigma ** 12 + \
           135135 * sigma ** 14


def gaussian_moment_15(mu: float, sigma: float = 1.) -> float:
    return mu ** 15 + \
           105 * mu ** 13 * sigma ** 2 + \
           1365 * mu ** 11 * 3 * sigma ** 4 + \
           5005 * mu ** 9 * 15 * sigma ** 6 + \
           6435 * mu ** 7 * 105 * sigma ** 8 + \
           3003 * mu ** 5 * 945 * sigma ** 10 + \
           455 * mu ** 3 * 10395 * sigma ** 12 + \
           15 * mu * 135135 * sigma ** 14


def gaussian_moment_16(mu: float, sigma: float = 1.) -> float:
    return mu ** 16 + \
           120 * mu ** 14 * sigma ** 2 + \
           1820 * mu ** 12 * 3 * sigma ** 4 + \
           8008 * mu ** 10 * 15 * sigma ** 6 + \
           12870 * mu ** 8 * 105 * sigma ** 8 + \
           8008 * mu ** 6 * 945 * sigma ** 10 + \
           1820 * mu ** 4 * 10395 * sigma ** 12 + \
           120 * mu ** 2 * 135135 * sigma ** 14 + \
           2027025 * sigma ** 16


"""Get the nt^h moment of a n-dim Gaussian (or normal distribution) centred at `mu`
with a standard deviation of `sigma`.

Taken from:
https://en.wikipedia.org/wiki/Normal_distribution#Moments
Can be generalised to any order using confluent hypergeometric functions of the second kind.

Another useful reference is:
http://www.randomservices.org/random/special/Normal.html

:param mu: the mean of the distribution
:param sigma: the standard deviation of the distribution
:param order: the order of the moment to get
:param weight: the total probability or mass of the normal distribution.  This is the zero^th
    moment be definition
"""
gaussian_moment = [
    np.vectorize(gaussian_moment_0),
    np.vectorize(gaussian_moment_1),
    np.vectorize(gaussian_moment_2),
    np.vectorize(gaussian_moment_3),
    np.vectorize(gaussian_moment_4),
    np.vectorize(gaussian_moment_5),
    np.vectorize(gaussian_moment_6),
    np.vectorize(gaussian_moment_7),
    np.vectorize(gaussian_moment_8),
    np.vectorize(gaussian_moment_9),
    np.vectorize(gaussian_moment_10),
    np.vectorize(gaussian_moment_11),
    np.vectorize(gaussian_moment_12),
    np.vectorize(gaussian_moment_13),
    np.vectorize(gaussian_moment_14),
    np.vectorize(gaussian_moment_15),
    np.vectorize(gaussian_moment_16),
]

gaussian_moment_derivatives = [
    np.vectorize(gaussian_moment_0_derivative),
    np.vectorize(gaussian_moment_1_derivative),
    np.vectorize(gaussian_moment_2_derivative),
    np.vectorize(gaussian_moment_3_derivative),
    np.vectorize(gaussian_moment_4_derivative),
    np.vectorize(gaussian_moment_5_derivative),
    np.vectorize(gaussian_moment_6_derivative),
    np.vectorize(gaussian_moment_7_derivative),
    np.vectorize(gaussian_moment_8_derivative),
]


def _check_gaussian_moments_input(
    mu: Union[numbers.Number, np.ndarray],
    sigma: Union[numbers.Number, np.ndarray],
    weight: numbers.Number,
):
    if isinstance(mu, np.ndarray):
        output_length = len(mu)
        input_length = output_length + 1  # Plus one for the weight
        if isinstance(sigma, np.ndarray):
            if isinstance(sigma, np.ndarray) and sigma.shape != mu.shape:
                raise ValueError(f'Mismatch between mu and sigma shapes: {mu.shape} != {sigma.shape}')
            input_length += len(sigma)
        else:
            input_length += 1  # Sigma scalar
    else:
        # Assume it's a scalar
        if isinstance(sigma, np.ndarray):
            raise TypeError('Scalar mu passed with vector sigma')

        input_length = 3
        output_length = 1

    if not isinstance(weight, numbers.Number):
        raise TypeError(f"Expecting weight to be a scalar, got '{weight.__class__.__name__}'")

    return input_length, output_length


def gaussian_geometric_moments(
    # pylint: disable=invalid-name
    max_order: int,
    mu: np.array,
    sigma: numbers.Number,
    weight: numbers.Number
) -> np.array:
    """
    Get the geometric moments for a 3D Gaussian

    :param mu: the position of the Gaussian distribution
    :param sigma: the standard deviation (scalar - same in all directions)
    :param max_order: the maximum order to calculate moments for
    :param weight: the total mass of the Gaussian (or equivalently total probability)
    :return: the 3d moment tensor
    """
    ubound = max_order + 1  # Calculate to max order (inclusive)

    moments = np.zeros((3, ubound))
    moments[:, 0] = 1.0  # 0^th order, mass is multiplied in at end

    for order in range(1, ubound):
        # Use weight 1 for now and then multiply later
        moments[:, order] = gaussian_moment[order](mu, sigma)

    moments_3d = np.empty((ubound,) * 3)
    # pylint: disable=invalid-name

    for p in range(ubound):
        for q in range(ubound):
            for r in range(ubound):
                moments_3d[p, q, r] = moments[0, p] * moments[1, q] * moments[2, r]

    moments_3d *= weight
    return moments_3d


def _to_array(value: Union[numbers.Number, np.array], shape):
    if isinstance(value, numbers.Number):
        sarray = np.empty(shape)
        sarray.fill(value)
        return sarray

    return value



# -*- coding: utf-8 -*-
"""Module for calculating Zernike moments"""

import collections
import logging
import itertools
import math
import numbers
import functools
import random
from typing import Union, Tuple, Iterator, Dict, Optional

import numpy as np
from numpy import ma
import scipy
import scipy.special

from . import base_moments
from . import functions
from . import geometric
from .utils import even, inclusive

_LOGGER = logging.getLogger(__name__)

# pylint: disable=invalid-name

__all__ = 'ZernikeMoments', 'ZernikeMomentCalculator', 'from_deltas', 'from_gaussians'


def from_deltas(max_order: int, positions: np.array, weights: Union[numbers.Number, np.array] = 1.) -> 'ZernikeMoments':
    """Create a set of Zernike moments from a collection of delta functions with optional weights

    :param max_order: the order of Zernike moments to calculate to
    :param positions: the Cartesian positions of the delta functions
    :param weights: the weights of the delta functions, can be a scalar or np.array of the same
        length as positions
    """
    _domain_check(positions)
    geom_moments = geometric.from_deltas(max_order, positions, weights)
    return from_geometric_moments(max_order, geom_moments)


def from_gaussians(
    max_order: int,
    positions: np.ndarray,
    sigmas: Union[numbers.Number, np.array] = 0.1,
    weights: Union[numbers.Number, np.array] = 0.1
) -> 'ZernikeMoments':
    """Create a set of Zernike moments from a collection of Gaussian functions with the given sigmas
    and weights

    :param max_order: the order of Zernike moments to calculate to
    :param positions: the Cartesian positions of the delta functions
    :param sigmas: the sigmas of the Gaussians, can be a scalar or np.array of the same length
        as positions
    :param weights: the weights of the delta functions, can be a scalar or np.array of the same
        length as positions
    """
    _domain_check(positions)
    geom_moments = geometric.from_gaussians(max_order, positions, sigmas, weights)
    return from_geometric_moments(max_order, geom_moments)


def from_geometric_moments(
    max_order: int, geom_moments: Union[geometric.GeometricMoments, np.array]
) -> 'ZernikeMoments':
    """Create a set of Zernike moments from a set of geometric moments

    :param max_order: the order of Zernike moments to calculate to
    :param geom_moments: the geometric moments
    """
    return ZernikeMomentCalculator(max_order)(geom_moments)


class ZernikeReconstructionQuery(base_moments.ReconstructionQuery):
    """A query object for Zernike moments that allows for faster reconstructions by caching the reconstruction grid"""

    def __init__(self, max_order: int, points: np.ndarray, moments: np.ndarray):
        super().__init__(max_order, points)
        self._moments = moments
        self._valid_idxs = None
        self._moments_in_domain = None

    @property
    def moments(self) -> np.ndarray:
        return self._moments

    @property
    def moments_in_domain(self) -> np.ndarray:
        if self._moments_in_domain is None:
            # Annoying indexing and unindexing, this returns a copy
            self._moments_in_domain = self._moments[:, :, :, self.valid_idxs][:, :, :, :, 0]
        return self._moments_in_domain

    @property
    def valid_idxs(self):
        """Get the indexes of points that are within the domain"""
        if self._valid_idxs is None:
            self._valid_idxs = ZernikeMoments._get_indices_in_domain(self.points)
        return self._valid_idxs


#: Helper tuple for working with Zernike moment indexes
ZernikeIndex = collections.namedtuple('ZernikeIndex', 'n l m')


class ZernikeMoments(base_moments.Moments):
    """A container class for calculated Zernike moments"""

    @classmethod
    def from_indexed(cls, indexed, max_order: int, dtype=None) -> 'ZernikeMoments':
        dtype = dtype or type(indexed[0, 0, 0])
        moments = cls(max_order, dtype=dtype)
        for (n, l, m) in moments.iter_indices():
            moments._set_moment(n, l, m, indexed[n, l, m], set_minus_m=False)
        return moments

    @classmethod
    def from_vector(cls, n_max: int, vec: np.array, redundant=True) -> 'ZernikeMoments':
        moms = ZernikeMoments(n_max, dtype=vec.dtype)

        for (n, l, m), value in zip(iter_indices(redundant=redundant), vec):
            if n > n_max:
                break
            moms._set_moment(n, l, m, value, set_minus_m=not redundant)  # pylint: disable=protected-access

        return moms

    @classmethod
    def rand(cls, n_max: int, num_range=(-5., 5.)):
        moms = ZernikeMoments(n_max)
        for nlm in moms.iter_indices(redundant=False):
            mom = random.uniform(*num_range)
            if nlm.m != 0:
                # Only l != 0 moments have a complex part
                mom += +1j * random.uniform(*num_range)
            moms[nlm] = mom

        return moms

    def __init__(self, n_max: int, dtype=complex):
        """Construct a Zernike moments object
        """
        self._max_n = n_max
        self._dtype = dtype

        moments = np.empty((n_max + 1, n_max + 1, 2 * n_max + 1), dtype=dtype)
        # Fill it with NaN so that it's easy to spot any values that haven't been set but should be
        moments.fill(np.nan)

        # Mask everything and then unmask the ones we want
        mask = np.ones(moments.shape, dtype=np.int)
        idx_array = np.empty(moments.shape, dtype=np.int)
        for i, indices in enumerate(iter_indices(n_max, redundant=True)):
            mask[indices] = 0
            idx_array[indices] = i

        self._moments = ma.array(moments, mask=mask)
        self._idx_array = ma.array(idx_array, mask=mask)

    def __eq__(self, other: 'ZernikeMoments'):
        """Test if another set of moments are equal to this one"""
        if not isinstance(other, ZernikeMoments):
            return False

        return np.all(self.vector == other.vector)

    @property
    def dtype(self):
        return self._dtype

    @property
    def array(self) -> np.ndarray:
        return self._moments

    def to_matrix(self):
        return self._moments

    @property
    def max_order(self) -> int:
        return self._max_n

    def get_builder(self, mask: 'Optional[ZernikeMoments]' = None):
        return ZernikeMomentsBuilder(self._max_n, mask)

    def get_mask(self) -> 'ZernikeMoments':
        moms = ZernikeMoments(self._max_n, dtype=object)
        moms.array.fill(None)
        return moms

    def randomise(self, num_range=(-5, 5), indices=(None, None, None)):
        for n, l, m in iter_indices_extended(indices, self._max_n, redundant=False):
            mom = random.uniform(*num_range)
            if m != 0:
                # Only l != 0 moments have a complex part
                mom += 1j * random.uniform(*num_range)
            self._set_moment(n, l, m, mom)

        return self

    @property
    def vector(self) -> np.ndarray:
        """Return this set of moments as a vector"""
        return self._moments.compressed()

    def __getitem__(self, item):
        return self.moment(*item)

    def __setitem__(self, key: Union[int, Tuple], value):
        self._set_moment(*key, value)

    def _set_moment(self, n: int, l: int, m: int, value, set_minus_m=True):
        """Set an individual moment.  This will automatically set the corresponding -m value"""
        if value is None:
            # Special case, usually happens if this is a mask
            self._moments[n, l, m] = None
            self._moments[n, l, -m] = None
        else:
            if m == 0:
                try:
                    if value.imag > 1e-9:
                        logging.warning(
                            f'Trying to set value of moment {n},{l},{m} to {value}, '
                            f'however these moment should always be real so discarding imaginary'
                        )
                    value = np.real(value)
                except AttributeError:
                    pass
            elif set_minus_m:
                # Set the -m counterpart directly
                self._moments[n, l, -m] = (-1)**m * value.conjugate()

            # This will be reached for all m values
            self._moments[n, l, m] = value

    def iter(self, redundant=True):
        """Iterate tuples containing the moment indices and value"""
        for indices in self.iter_indices(redundant=redundant):
            yield indices, self[indices]

    def iter_indices(self, redundant=True):
        yield from iter_indices(max_order=self._max_n, redundant=redundant)

    def linear_index(self, index: base_moments.Index) -> int:
        """Given a triple of Zernike moment indices this will return the corresponding linearly ordered index"""
        return self._idx_array[index]

    @staticmethod
    def num_moments(max_order: int, redundant=True) -> int:
        """Get the total number of Zernike moments up to the maximum order"""
        return sum(1 for _ in iter_indices(max_order, redundant=redundant))

    def moment(self, n: int, l: int, m: int) -> complex:
        """Get the n, l, m^th moment"""
        return self._moments[n, l, m]

    def value_at(self, x: np.array, order: int = None) -> float:
        """Reconstruct the value at x from the moments

        :param x: the point to get the value at
        :param order: the maximum order to go up to (defaults to the order of these moments)
        """
        order = order or self._max_n
        return self.reconstruct(self.create_reconstruction_query(x, order), order, zero_outside_domain=False)

    def reconstruct(self, query: ZernikeReconstructionQuery, order=None, zero_outside_domain=True):
        """Given a reconstruction query this will return the values of the function at the specified grid values"""
        order = order if order is not None else query.max_order
        moments = query.moments

        if len(query.points.shape) > 1:
            values = np.zeros(query.points.shape[0], dtype=self.dtype)
        else:
            values = 0.

        if zero_outside_domain:
            # Use only reconstruction moments from within the domain
            moments = query.moments_in_domain
            calculated_values = np.zeros(query.valid_idxs.size, dtype=self.dtype)
        else:
            calculated_values = values

        for n, l, m in iter_indices(max_order=order, redundant=False):
            omega = self.moment(n, l, m)
            z = sum_chi_nlm(n, l, m, moments)
            calculated_values += omega * z

            if m != 0:
                # Now do the symmetric -m part
                z_conj = (-1)**m * z.conjugate()

                calculated_values += z_conj * self.moment(n, l, -m)

        if zero_outside_domain:
            np.put(values, query.valid_idxs, calculated_values)

        return values.real

    def visualise(
        self, viewer='plotly_widget', query: ZernikeReconstructionQuery = None, max_order=None, num_points=31
    ):
        if viewer.startswith('plotly_widget'):
            volume_plot = viewer.endswith('volume')

            # Only do imports here so some of these tools aren't needed for the rest of milad
            import ipywidgets
            import plotly.graph_objects as go
            from milad import plotlytools

            if query is None:
                grid_points = self.get_grid(num_points, restrict_to_domain=False)
                query = self.create_reconstruction_query(grid_points, self.max_order)
            else:
                grid_points = query.points

            grid_values = self.reconstruct(query, max_order, zero_outside_domain=True)

            default_isomin = (grid_values.max() - grid_values.min()) * 0.6 + grid_values.min()
            default_isomax = grid_values.max()

            if volume_plot:
                data = plotlytools.volume(
                    grid_points,
                    grid_values,
                    isomin=grid_values.mean(),
                    opacity=0.1,
                    autocolorscale=True,
                )
            else:
                data = plotlytools.isosurface(
                    grid_points,
                    grid_values,
                    isomin=grid_values.mean(),
                    opacity=0.1,
                    autocolorscale=True,
                )

            fig = go.Figure(data=[data])
            widget = go.FigureWidget(fig)

            default_opacity = 0.05 if volume_plot else 0.4

            @ipywidgets.interact(
                opacity=(0., 1., 0.01),
                isomin=(grid_values.min(), grid_values.max(), 0.5),
                isomax=(grid_values.min(), grid_values.max(), 0.5)
            )
            def update(opacity=default_opacity, isomin=default_isomin, isomax=default_isomax):  # pylint: disable=unused-variable
                with widget.batch_update():
                    isosurface = widget.data[0]
                    isosurface.opacity = opacity
                    isosurface.isomin = isomin
                    isosurface.isomax = isomax

            return widget

        raise ValueError(f"Unknown viewer '{viewer}'")

    @classmethod
    def create_reconstruction_query(cls, points: np.ndarray, max_order: int) -> ZernikeReconstructionQuery:
        """Create a query object that can be passed to self.value_at to reconstruct the Zernike functions
        at the given query points.  This is useful if the same point (or grid of points) is used multiple
        times as this precomputation can be done once and reused.

        :param points: the point to get the value at
        :param max_order: the maximum order to go up to (defaults to the order of these moments)
        """
        return ZernikeReconstructionQuery(max_order, points, cls._get_geometric_moments(points, max_order))

    @classmethod
    def _get_geometric_moments(cls, point: np.array, order: int) -> np.ndarray:
        """Get geometric moments from a point or set of points.  Used in reconstruction."""
        if len(point.shape) == 2:
            moms = [geometric.from_deltas(order, [pt]).to_matrix() for pt in point]
            moments = np.empty(list(moms[0].shape) + [len(moms)])
            for idx, entry in enumerate(moms):
                moments[:, :, :, idx] = entry
        else:
            moments = geometric.from_deltas(order, [point])

        return moments

    @staticmethod
    def _get_indices_in_domain(points: np.ndarray) -> np.ndarray:
        """Calculate the lengths squared and get the corresponding indexes"""
        length_sq = (points**2).sum(axis=1)
        return np.argwhere(length_sq < 1)


class ZernikeMomentsBuilder(functions.Function):
    """This builder takes a vector of real values and constructs a set of complex Zernike moments."""
    input_type = np.ndarray
    supports_jacobian = True

    def __init__(self, n_max: int, mask: Optional[ZernikeMoments] = None):
        super().__init__()
        self._n_max = n_max
        self._mask = mask
        self._indices = None

    def apply_mask(self, moms: ZernikeMoments):
        """Given a set of moments this will set any values present from the mask"""
        if self._mask is None:
            raise RuntimeError('Cannot apply mask as, mask is None')
        for zernike_index, value in self._mask.iter(redundant=False):
            if value is not None:
                moms[zernike_index] = value

    @property
    def inverse(self) -> 'ZernikeMomentsBuilder.Inverse':
        return ZernikeMomentsBuilder.Inverse(self)

    def evaluate(self, state: np.ndarray, get_jacobian=False):
        indices_list = tuple(self.iter_indices())
        if len(state) != len(indices_list):
            raise ValueError(
                f'Got input vector of length {len(state)} but was expecting one of length {len(indices_list)}'
            )

        # Create the moments and optionally a Jacobian
        moms = ZernikeMoments(self._n_max)
        moms.array.fill(0.)
        if self._mask is not None:
            self.apply_mask(moms)
        jac = np.zeros((len(moms), len(indices_list)), dtype=complex) if get_jacobian else None

        # Let's create the complex moments
        for idx, (zernike_idx, num_type) in enumerate(indices_list):
            moms[zernike_idx] += num_type * state[idx]

            if get_jacobian:
                n, l, m = zernike_idx
                lin_idx = moms.linear_index(zernike_idx)
                jac[lin_idx, idx] += 1 * num_type
                if zernike_idx.m > 0:
                    lin_idx = moms.linear_index((n, l, -m))
                    jac[lin_idx, idx] += (-1)**m * (1 if num_type == 1. else -1j)

        if get_jacobian:
            return moms, jac

        return moms

    def iter_indices(self):
        # Skip all the ones that are masked
        if self._indices is None:
            # Lazily create the indices
            self._indices = tuple(self._iter_indices(self._n_max))

        for zernike_index, num_type in self._indices:
            if self._mask is None or self._mask[zernike_index] is None:
                yield zernike_index, num_type

    @classmethod
    def _iter_indices(cls, n_max: int):
        # First deal with m = 0 as these are all real
        for n in inclusive(0, n_max):
            for l in inclusive(n):
                if not even(n - l):
                    continue

                yield ZernikeIndex(n, l, 0), 1.

        # Now deal with m > 0, these have both a real and imaginary part
        for num_type in (1., 1j):
            for n in inclusive(n_max):
                for l in inclusive(n):
                    if not even(n - l):
                        continue

                    for m in inclusive(1, l, 1):
                        yield ZernikeIndex(n, l, m), num_type

    class Inverse(functions.Function):
        """Turn a set a Zernike moments into a vector"""
        supports_jacobian = False

        def __init__(self, forward: 'ZernikeMomentsBuilder'):
            super().__init__()
            self._forward = forward

        @property
        def inverse(self) -> 'ZernikeMomentsBuilder':
            return self._forward

        def evaluate(# pylint: disable=unused-argument
                self, state: ZernikeMoments, *, get_jacobian=False) -> np.ndarray:
            moms = []
            for indices, num_type in self._forward.iter_indices():
                mom = state[indices]
                moms.append(mom.real if num_type == 1. else mom.imag)

            return np.array(moms)


class ZernikeMomentCalculator(base_moments.MomentsCalculator):
    """Calculate Zernike moments.

    Takes as input geometric moments or any state vector that is a valid input to GeometricMomentsCalculator
    """
    output_type = ZernikeMoments
    supports_jacobian = True
    dtype = complex
    input_type = geometric.GeometricMoments, np.ndarray, functions.Features

    def __init__(self, max_order: int):
        super().__init__()
        self._max_order = max_order
        self._geometric_moments_calculator = geometric.GeometricMomentsCalculator(max_order)
        self._jacobian = None
        self._chi = None  # The omega_nl_m prefactors

    def evaluate(self,
                 state: functions.State,
                 *,
                 get_jacobian=False) -> Union[ZernikeMoments, Tuple[ZernikeMoments, np.ndarray]]:
        if isinstance(state, geometric.GeometricMoments):
            geom_moments = state
            geom_jac = np.eye(state.size)
        else:
            geom_moments = self._geometric_moments_calculator(state, get_jacobian)
            if get_jacobian:
                geom_moments, geom_jac = geom_moments

        moments = ZernikeMoments(self._max_order, dtype=object if geom_moments.dtype == object else complex)

        # Get the moments themselves from polynomials of geometric moments
        for (n, l, m), poly in self.chi.items():
            moments[n, l, m] = poly.evaluate(geom_moments)

        if get_jacobian:
            jac = np.matmul(self._get_jacobian(), geom_jac)
            return moments, jac

        return moments

    def _get_jacobian(self):
        if self._jacobian is None:
            # Calculate the first time
            self._jacobian = get_jacobian_wrt_geom_moments(self._max_order, redundant=True)

        return self._jacobian

    @property
    def chi(self) -> Dict[base_moments.Index, base_moments.MomentsPolynomial]:
        """Get the chi_nlm dictionary.  Indices are moment indices and values are polynomials"""
        if self._chi is None:

            class Tracker:
                """Keeps track of the moments evaluated by omega_nl_m"""

                def __init__(self, max_order):
                    self._max_order = max_order

                def __getitem__(self, item):
                    mtx = np.zeros((self._max_order + 1, self._max_order + 1, self._max_order + 1), dtype=complex)
                    mtx[item] = 1.
                    return mtx

            chi = collections.OrderedDict()
            for n, l, m in iter_indices(self._max_order, redundant=False):
                used = omega_nl_m(n, l, m, Tracker(self._max_order))
                indexes = np.argwhere(used)
                poly = base_moments.MomentsPolynomial()

                for index in indexes:
                    index = tuple(index)
                    poly.append(base_moments.Product(used[index], (index,)))

                chi[(n, l, m)] = poly

            self._chi = chi

        return self._chi

    def create_random(self, max_order: int = None) -> ZernikeMoments:
        max_order = max_order or self._max_order
        return ZernikeMoments.rand(max_order)


def get_jacobian_wrt_geom_moments(max_order: int, redundant=True):
    """Get the Jacobian of the Zernike moments wrt Geometric moments as input"""

    # Calculate
    class DerivativeTracker:

        def __init__(self, max_ord):
            self._max_order = max_ord
            self._num_coeffs = geometric.GeometricMoments.num_moments(max_ord)

        def __getitem__(self, item: base_moments.Index):
            if not isinstance(item, tuple):
                raise TypeError(item)

            vector = np.zeros(self._num_coeffs)
            try:
                vector[geometric.linear_index(self._max_order, item)] = 1.
            except IndexError:
                raise IndexError(f'index {item} out of bounds for geometric moments of order {self._max_order}')
            return vector

    tracker = DerivativeTracker(max_order)
    num_moments = ZernikeMoments.num_moments(max_order, redundant=redundant)  # Number of zernike moments
    num_geometric_moments = geometric.GeometricMoments.num_moments(max_order)
    jacobian = np.zeros((num_moments, num_geometric_moments), dtype=complex)

    for idx, (n, l, m) in enumerate(iter_indices(max_order, redundant=redundant)):
        if m < 0:
            # These are taken care of during the positive m iteration
            continue

        vec = omega_nl_m(n, l, m, tracker)
        jacobian[idx, :] = vec

        if redundant and m != 0:
            minus_m = (-1)**m * vec.conjugate()
            minus_m_idx = linear_index((n, l, -m))
            jacobian[minus_m_idx, :] = minus_m

    return jacobian


@functools.lru_cache(maxsize=None)
def factorial(n: int) -> int:
    return scipy.special.factorial(n, exact=True)


@functools.lru_cache(maxsize=None)
def binomial_coeff(n: int, k: int) -> int:
    return scipy.special.comb(n, k, exact=True)


@functools.lru_cache(maxsize=None)
def c_l_m(l: int, m: int) -> float:
    """Calculate the normalisation factor"""
    if m < 0:
        # make use of symmetry c_l^m == c_l^-m
        return c_l_m(l, -m)

    return ((2 * l + 1) * factorial(l + m) * factorial(l - m))**0.5 / factorial(l)


@functools.lru_cache(maxsize=None)
def q_kl_nu(k: int, l: int, nu: int) -> float:
    """Calculate the Zernike geometric moment conversion factors"""
    return (-1) ** k / 2 ** (2 * k) * \
           ((2 * l + 4 * k + 3) / 3) ** 0.5 * \
           binomial_coeff(2 * k, k) * \
           (-1) ** nu * \
           (binomial_coeff(k, nu) * binomial_coeff(2 * (k + l + nu) + 1, 2 * k)) / \
           binomial_coeff(k + l + nu, k)


def sum_chi_nlm(n: int, l: int, m: int, geom_moments: np.ndarray) -> Union[complex, np.array]:
    """Calculate the Zernike geometric moment conversion factors"""
    return c_l_m(l, m) * 2**(-m) * sum1(n=n, l=l, m=m, geom_moments=geom_moments)


def sum1(n: int, l: int, m: int, geom_moments: np.ndarray) -> Union[complex, np.array]:
    """Calculate the Zernike geometric moment conversion factors"""
    k = int(math.floor((n - l) / 2))  # n - l is always even
    total = 0.
    for nu in inclusive(k):
        total += q_kl_nu(k, l, nu) * sum2(n=n, l=l, m=m, nu=nu, geom_moments=geom_moments)

    return total


def sum2(n: int, l: int, m: int, nu: int, geom_moments: np.ndarray) -> Union[complex, np.array]:
    total = 0.
    for alpha in inclusive(nu):
        total += binomial_coeff(nu, alpha) * \
                 sum3(n=n, l=l, m=m, nu=nu, alpha=alpha, geom_moments=geom_moments)
    return total


def sum3(n: int, l: int, m: int, nu: int, alpha: int, geom_moments: np.ndarray) -> Union[complex, np.array]:
    total = 0.
    for beta in inclusive(nu - alpha):
        total += binomial_coeff(nu - alpha, beta) * \
                 sum4(n=n, l=l, m=m, nu=nu, alpha=alpha, beta=beta, geom_moments=geom_moments)
    return total


def sum4(n: int, l: int, m: int, nu: int, alpha: int, beta: int, geom_moments: np.ndarray) -> Union[complex, np.array]:
    total = 0.
    for u in inclusive(m):
        total += (-1) ** (m - u) * binomial_coeff(m, u) * 1j ** u * \
                 sum5(n=n, l=l, m=m, nu=nu, alpha=alpha, beta=beta, u=u,
                      geom_moments=geom_moments)
    return total


def sum5(n: int, l: int, m: int, nu: int, alpha: int, beta: int, u: int,
         geom_moments: np.ndarray) -> Union[complex, np.array]:
    total = 0.
    for mu in inclusive(int(math.floor((l - m) / 2.))):
        total += (-1) ** mu * 2 ** (-2 * mu) * binomial_coeff(l, mu) * \
                 binomial_coeff(l - mu, m + mu) * \
                 sum6(n=n, l=l, m=m, nu=nu, alpha=alpha, beta=beta, u=u, mu=mu,
                      geom_moments=geom_moments)
    return total


def sum6(n: int, l: int, m: int, nu: int, alpha: int, beta: int, u: int, mu: int,
         geom_moments: np.ndarray) -> Union[complex, np.array]:
    total = 0.
    for v in inclusive(mu):
        r = 2 * (v + alpha) + u
        s = 2 * (mu - v + beta) + m - u
        t = 2 * (nu - alpha - beta - mu) + l - m

        if r + s + t > n:
            continue

        total += binomial_coeff(mu, v) * geom_moments[r, s, t]

    return total


def assert_valid(n: int, l: int, m: int):
    if not n >= 0:
        raise ValueError('n must be a positive integer')
    if not 0 <= l <= n:
        raise ValueError(f'l must be 0 <= l <= n, got n={n}, l={l}')
    if not -l <= m <= l:
        raise ValueError(f'm must be in the range -{l} <= m <= {l}, got {m}')
    if not even(n - l):
        raise ValueError('n - l must be even')


def omega_nl_m(n: int, l: int, m: int, geom_moments: np.array) -> complex:
    """Given a set of geometric moments this function will compute the corresponding Zernike moments
    """
    assert_valid(n, l, m)
    if m < 0:
        # Symmetry relation
        return (-1)**(-m) * omega_nl_m(n, l, -m, geom_moments).conjugate()

    return 3. / (4. * np.pi) * sum_chi_nlm(n, l, m, geom_moments).conjugate()


def iter_indices(max_order: int = None, redundant=False) -> Iterator[ZernikeIndex]:
    """Iterate over Zernike function indices in lexicographic order.

    If redundant is True then all valid indices will be generated including those where there is a
    symmetry relation e.g.: o_22^-2 = (-1)^2 o22^2.

    :param max_order: order to provide indices up to, if None will generate all indices
    :param redundant: if True include redundant indices
    """
    upper = itertools.count() if max_order is None else inclusive(max_order)
    for n in upper:
        for l in inclusive(n):
            if not even(n - l):
                continue

            for m in inclusive(0, l):
                yield ZernikeIndex(n, l, m)

            # negative m indices go AFTER positive because we rely on the wrapping-around of negative indices
            # to make these work they are actually at the end of the sequence of m values
            if redundant:
                for m in inclusive(-l, -1, 1):
                    yield ZernikeIndex(n, l, m)


def iter_indices_extended(key: Tuple[int, Tuple], max_order: int, redundant=False) -> Iterator[ZernikeIndex]:
    if isinstance(key, int):
        # Assume the caller is treating us like a vector
        n, l, m = triple_index(key)
    else:
        # Assume tuple
        n, l, m = key

    if n is None or l is None or m is None:
        # Define the range of n values
        if n is None:
            n_iter = inclusive(0 if l is None else l, max_order)
        else:
            n_iter = inclusive(n, n)

        for n_idx in n_iter:
            # Define the range of l values
            if l is None:
                l_iter = inclusive(0, n_idx)
            else:
                l_iter = inclusive(l, l)

            for l_idx in l_iter:
                if not even(n_idx - l_idx):
                    continue

                # Define the range of m values
                if m is None:
                    m_start = -l_idx if redundant else 0
                    m_iter = inclusive(m_start, l_idx)
                else:
                    m_iter = inclusive(m, m)

                for m_idx in m_iter:
                    yield ZernikeIndex(n_idx, l_idx, m_idx)

    else:
        # All direct indices
        yield ZernikeIndex(n, l, m)


def triple_index(linear: int, redundant=True) -> ZernikeIndex:
    """Get the triple index from the given lexicographic index"""
    entry = None
    for entry, _ in zip(iter_indices(redundant=redundant), range(linear + 1)):
        pass
    return entry


@functools.lru_cache(maxsize=256)
def linear_index(index: base_moments.Index, redundant=True) -> int:
    """Given a triple of zernike function indices this will return a lexicographically ordered
    integer"""
    if not redundant and index[2] < 0:
        raise ValueError('m value cannot be negative')

    for linear, triple in enumerate(iter_indices(redundant=redundant)):
        if triple == index:
            return linear

    # pylint: disable=inconsistent-return-statements
    assert False, 'Should never reach here'


def _domain_check(positions: np.array):
    """Check that a given set of positions are within the domain for which Zernike moments are
    defined i.e. that |r| <= 1."""
    if positions.dtype == object:
        return

    for idx, pos in enumerate(positions):
        if np.dot(pos, pos) > 1.:
            _LOGGER.warning(
                'Delta function %i is outside of the domain of Zernike basis functions '
                'with are defined within the unit ball', idx
            )



# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt
import numpy as np

from . import neuralnetwork

__all__ = ('TrainingMonitor',)


class TrainingMonitor:
    """Class that supports monitoring and plotting results of a neural network training procedure"""

    class PlottingDataset:

        def __init__(self, axes, scatter_kwargs: dict = None):
            self._axes = axes
            self._last_plt = None
            self._values = []
            self._scatter_kwargs = scatter_kwargs or {}

        def append(self, value):
            self._values.append(value)

        def replot(self):
            if self._last_plt is not None:
                self._last_plt.remove()
            self._last_plt = self._axes.scatter(list(range(len(self._values))), self._values, **self._scatter_kwargs)

    def __init__(self, validation_data: neuralnetwork.FittingData = None):
        fig = plt.figure(figsize=(12, 10))
        energy_axis = fig.add_subplot(111)
        energy_axis.set_yscale('log')
        plt.ion()

        self.fig = fig
        self.energy_axis = energy_axis

        self._energy_training = TrainingMonitor.PlottingDataset(
            self.energy_axis,
            scatter_kwargs=dict(
                edgecolors='tab:orange', alpha=0.6, marker='o', facecolors='none', label='Energy (training)'
            )
        )
        self._energy_validation = TrainingMonitor.PlottingDataset(
            self.energy_axis, scatter_kwargs=dict(c='tab:orange', alpha=0.6, label='Energy (validation)')
        )

        forces_axis = energy_axis.twinx()
        forces_axis.set_yscale('log')

        self._force_training = TrainingMonitor.PlottingDataset(
            forces_axis,
            scatter_kwargs=dict(edgecolors='tab:blue', alpha=0.6, facecolors='none', label='Force (training)')
        )
        self._force_validation = TrainingMonitor.PlottingDataset(
            forces_axis, scatter_kwargs=dict(c='tab:blue', alpha=0.6, label='Force (validation)')
        )

        self._validation_data = validation_data

    def progress_callaback(self, network, _step, _training, loss):
        # Calculate MSEs
        training_rmsd = loss.energy.cpu().item()**0.5
        self._energy_training.append(training_rmsd)
        self._energy_training.replot()

        if loss.force is not None:
            self._force_training.append(loss.force.cpu().item()**0.5)
            self._force_training.replot()

        if self._validation_data:
            # Calculate the validation loss
            validation_loss = network.loss(self._validation_data)
            validation_rmsd = validation_loss.energy.cpu().item()**0.5
            self._energy_validation.append(validation_rmsd)
            self._energy_validation.replot()

            if validation_loss.force is not None:
                force_rmsd = validation_loss.force.cpu().item()**0.5
                self._force_validation.append(force_rmsd)
                self._force_validation.replot()

        self.fig.canvas.draw()

    def plot_energy_comparison(self, network: neuralnetwork.NeuralNetwork, *fitting_data: neuralnetwork.FittingData):
        fig = plt.figure(figsize=(8, 8))
        axis = fig.gca()

        minimum, maximum = np.inf, -np.inf

        training_data = None if not fitting_data else fitting_data[0]

        if training_data:
            predictions = network.make_prediction(training_data)
            known_energies = training_data.get_normalised_energies().cpu().detach().numpy()
            axis.scatter(known_energies, predictions.get_normalised_energies().cpu().detach().numpy(), c='r', alpha=0.3)
            minimum = min(minimum, known_energies.min())
            maximum = max(maximum, known_energies.max())

        for entry in fitting_data:
            # Any additional datasets the user wants to plot
            predictions = network.make_prediction(entry)
            known_energies = entry.get_normalised_energies().cpu().detach().numpy()
            axis.scatter(known_energies, predictions.get_normalised_energies().cpu().detach().numpy(), alpha=0.4)
            minimum = min(minimum, known_energies.min())
            maximum = max(maximum, known_energies.max())

        if self._validation_data:
            predictions = network.make_prediction(self._validation_data)
            known_energies = self._validation_data.get_normalised_energies().cpu().detach().numpy()
            axis.scatter(known_energies, predictions.get_normalised_energies().cpu().detach().numpy(), c='b', alpha=0.8)
            minimum = min(minimum, known_energies.min())
            maximum = max(maximum, known_energies.max())

        if minimum != np.inf:
            axis.plot((minimum, maximum), (minimum, maximum), 'k-', lw=2, c='black')

        return fig

    def plot_energy_deviation_histogram(self, network, *fitting_data: neuralnetwork.FittingData):
        fig = plt.figure(figsize=(10, 10))
        axis = fig.gca()
        axis.set_ylabel('No. of structures')

        all_datasets = [self._validation_data, *fitting_data]
        for data_set in all_datasets:
            target_energies = data_set.get_normalised_energies().cpu().detach().numpy()
            predicted_energies = network.make_prediction(data_set).get_normalised_energies().cpu().detach().numpy()
            differences = target_energies - predicted_energies
            axis.hist(differences, 100)

        return fig



# -*- coding: utf-8 -*-
"""
Module containing functions and objects related to manipulating collections of atoms
"""
import copy
import logging
import random
from typing import Optional, Type, Tuple, Union, List

import numpy as np
from scipy.spatial.distance import pdist

from . import functions
from . import generate

_LOGGER = logging.getLogger(__name__)

__all__ = 'AtomsCollection', 'random_atom_collection_in_sphere'


class AtomsCollection(functions.PlainState):
    """A collection of atomic positions along with a numbers for each of their species.  Stored as:

    [x_1, y_1, z_1, x_2, y_2, z_2, ..., x_N, y_N, z_N, s_1, s_2, ..., S_N]

    where s_i are the numbers labelling the species
    """

    @staticmethod
    def num_atomic_properties() -> int:
        """Returns the number of atomic properties i.e. variables that make up each atom.  Currently these are
        the x, y, z positions and the atomic specie but this may change."""
        return 4

    @staticmethod
    def contains_num_atoms(state: functions.StateLike) -> int:
        length = len(state)
        if (length % AtomsCollection.num_atomic_properties()) != 0:
            raise ValueError('AtomsCollection state vectors must be a multiple of 4')
        return int(length / AtomsCollection.num_atomic_properties())

    @staticmethod
    def total_length(num_atoms: int) -> int:
        return AtomsCollection.num_atomic_properties() * num_atoms

    def __init__(self, num: int, positions: np.array = None, numbers: np.array = None, dtype=None):
        super().__init__(self.total_length(num), dtype=dtype)

        self._num_atoms = num
        if positions is not None:
            self.positions = positions
        if numbers is not None:
            self.numbers = numbers

    def __str__(self) -> str:
        return f'{len(self)} {str(self.numbers)}'

    def copy(self) -> 'AtomsCollection':
        atoms = AtomsCollection(self._num_atoms, dtype=self.dtype)
        np.copyto(atoms.vector, self.vector)
        return atoms

    def linear_pos_idx(self, atom_idx: int) -> slice:
        """Get the slice of the linear vector containing the position of the atom with the given
        index"""
        if atom_idx < 0 or atom_idx >= self._num_atoms:
            raise IndexError(atom_idx)
        return slice(3 * atom_idx, 3 * (atom_idx + 1))

    def linear_number_idx(self, atom_idx: int) -> int:
        """Get the index of the weight of the atom with the given index"""
        if atom_idx < 0 or atom_idx >= self._num_atoms:
            raise IndexError(atom_idx)

        return 3 * self._num_atoms + atom_idx

    @property
    def num_atoms(self):
        return self._num_atoms

    @property
    def positions(self) -> np.array:
        positions = self.vector[:3 * self._num_atoms]
        positions.shape = (self._num_atoms, 3)
        return positions

    @positions.setter
    def positions(self, new_positions: np.array):
        positions = self.vector[:3 * self._num_atoms]
        positions.shape = (self._num_atoms, 3)
        positions[:, :] = new_positions

    @property
    def numbers(self) -> np.array:
        return self.vector[3 * self._num_atoms:]

    @numbers.setter
    def numbers(self, new_numbers: np.array):
        self._array[3 * self._num_atoms:] = new_numbers

    def get_mask(self, fill=None):
        """Get a mask for this atom collection.  Typically used during optimisation."""
        mask = AtomsCollection(self.num_atoms, dtype=object)
        mask.vector.fill(fill)
        return mask

    def get_builder(self, mask: 'Optional[AtomsCollection]' = None):
        return AtomsCollectionBuilder(self._num_atoms, mask=mask)


class AtomsCollectionBuilder(functions.Function):
    """Take a vector as an input and build the corresponding atoms collection from it"""
    input_type = np.ndarray
    output_type = AtomsCollection
    supports_jacobian = True

    def __init__(self, num_atoms: int, mask: AtomsCollection = None):
        super().__init__()
        self._num_atoms = num_atoms
        if mask is None:
            self._mask = AtomsCollection(num_atoms, dtype=object)
            self._mask.positions[:] = None
            self._mask.numbers[:] = None
        else:
            self._mask = mask

    @property
    def num_atoms(self) -> int:
        """Get the number of atoms supported by this builder"""
        return self._num_atoms

    @property
    def mask(self) -> AtomsCollection:
        """Return the mask of fixed values (free values will have None entries)"""
        return self._mask

    @property
    def positions(self) -> np.ndarray:
        positions = self._mask[:3 * self._num_atoms]
        positions.shape = (self._num_atoms, 3)
        return positions

    @positions.setter
    def positions(self, value: np.ndarray):
        self._mask[:3 * self._num_atoms] = value.reshape(3 * self._num_atoms)

    @property
    def numbers(self) -> np.array:
        return self._mask[3 * self._num_atoms:]

    @numbers.setter
    def numbers(self, value: np.ndarray):
        self._mask[3 * self._num_atoms:] = value

    @property
    def input_length(self) -> int:
        return (self._mask.vector == None).sum()  # pylint: disable=singleton-comparison

    @property
    def inverse(self) -> 'AtomsCollectionBuilder.Inverse':
        return AtomsCollectionBuilder.Inverse(self)

    def output_length(self, _in_state: functions.State) -> int:
        return 4 * self._num_atoms

    def apply_mask(self, atoms: AtomsCollection):
        """Given an atom collection this will set any values specified in the mask"""
        indices = tuple(np.argwhere(self._mask.vector != None).reshape(-1))  # pylint: disable=singleton-comparison
        if len(indices) != 0:
            functions.copy_to(atoms.vector, indices, self._mask.vector, indices)

    def evaluate(self,
                 state: functions.State,
                 *,
                 get_jacobian=False) -> Union[AtomsCollection, Tuple[AtomsCollection, np.ndarray]]:
        vector = functions.get_bare(state)
        if len(vector) != self.input_length:
            raise ValueError(f'Expected input of length {self.input_length} but got {len(vector)}')

        atoms = AtomsCollection(self._num_atoms)
        self.apply_mask(atoms)

        # Get the unmasked indices
        indices = np.argwhere(self._mask.vector == None).reshape(-1)  # pylint: disable=singleton-comparison
        functions.copy_to(atoms.vector, indices, vector)

        if get_jacobian:
            jacobian = np.zeros((self.output_length(state), len(state)))
            for in_idx, out_idx in enumerate(indices):
                jacobian[out_idx, in_idx] = 1.

            return atoms, jacobian

        return atoms

    class Inverse(functions.Function):
        input_type = AtomsCollection
        supports_jacobian = False

        def __init__(self, builder: 'AtomsCollectionBuilder'):
            super().__init__()
            self._builder = builder

        @property
        def inverse(self) -> 'AtomsCollectionBuilder':
            return self._builder

        def output_length(self, _in_state: AtomsCollection) -> int:
            return self._builder.input_length

        def evaluate(self, state: AtomsCollection, *, get_jacobian=False) -> np.ndarray:
            output_length = self.output_length(state)
            if self.output_length(state) != output_length:
                raise ValueError(f'Expected input of length {output_length} but got {state.vector.size}')

            indices = np.argwhere(self._builder.mask.vector == None)  # pylint: disable=singleton-comparison
            return state.vector[indices].reshape(-1)


class FeatureMapper(functions.Function):
    """Map a collection of atoms onto a generic feature type such as Delta functions or Gaussians"""
    input_type = AtomsCollection
    output_type = functions.Features
    supports_jacobian = True

    def __init__(
        self,
        type: Type[functions.Feature] = functions.WeightedDelta,  # pylint: disable=redefined-builtin
        kwargs: dict = None,
        map_species_to: Union[int, str] = None
    ):
        """
        :param type: the feature type
        :param kwargs: dictionary of keyword arguments to pass to construct the features with
        :param map_species_to: optionally map atomic numbers to this index of the feature function vector
        """
        super().__init__()
        self._feature_type = type
        self._map_species_to = self._get_species_map_idx(type, map_species_to)
        self._feature_kwargs = kwargs or {}
        self._inverse = self.Inverse(self)

    def __repr__(self) -> str:
        desc = [self._feature_type.__name__]
        if self._feature_kwargs:
            desc.append(f'({self._feature_type})')
        desc.append(f' (species=>{self._map_species_to})')
        return ''.join(desc)

    @property
    def inverse(self):
        return self._inverse

    @property
    def map_species_to(self) -> Optional[str]:
        return self._map_species_to

    @staticmethod
    def _get_species_map_idx(feature_type: Type[functions.Feature], map_to: Union[int, str]) -> Optional[int]:
        if isinstance(map_to, int):
            return map_to
        if isinstance(map_to, str):
            return getattr(feature_type, map_to)
        if map_to is None:
            return None

        raise TypeError(map_to)

    def output_length(self, in_state: AtomsCollection) -> int:
        return in_state.num_atoms * self._feature_type.LENGTH

    def evaluate(self, atoms: AtomsCollection, *, get_jacobian=False) -> functions.Features:  # pylint: disable=arguments-differ
        features = functions.Features()

        jac = np.zeros((self.output_length(atoms), len(atoms))) if get_jacobian else None

        # Create the features one by one and add them to the features vector
        idx = 0  # Keep track of where we are in the features vector
        for atom_idx, (pos, specie) in enumerate(zip(atoms.positions, atoms.numbers)):
            kwargs = self._feature_kwargs.copy()
            kwargs['pos'] = pos
            feature = self._feature_type(**kwargs)
            if self._map_species_to:
                feature.vector[self._map_species_to] = specie

            if get_jacobian:
                # The structure of each feature is assumed to be [x, y, z, ..., s, ...]
                # where we optionally map the species to the index s
                pos_idx = atoms.linear_pos_idx(atom_idx).start
                for i in range(0, 3):
                    jac[idx + i, pos_idx + i] = 1.

                if self._map_species_to:
                    jac[idx + self._map_species_to, atoms.linear_number_idx(atom_idx)] = 1.

            features.add(feature)
            idx += len(feature)

        if get_jacobian:
            return features, jac

        return features

    class Inverse(functions.Function):
        input_type = functions.Features
        output_type = AtomsCollection
        supports_jacobian = False

        def __init__(self, mapper: 'FeatureMapper'):
            super().__init__()
            self._mapper = mapper

        @staticmethod
        def output_length(features: functions.Features) -> int:
            return AtomsCollection.total_length(len(features.features))

        def evaluate(self, features: functions.Features, get_jacobian=False):  # pylint: disable=arguments-differ
            atoms_collection = AtomsCollection(len(features.features))

            for idx, feature in enumerate(features.features):
                atoms_collection.positions[idx] = feature.pos
                if self._mapper.map_species_to:
                    atoms_collection.numbers[idx] = getattr(feature, self._mapper.map_species_to)

            return atoms_collection


class ScalePositions(functions.Function):
    """Scale the atomic system by a fixed scale factor"""

    input_type = AtomsCollection
    output_type = AtomsCollection
    supports_jacobian = True

    def __init__(self, scale_factor: float):
        super().__init__()
        self._scale_factor = scale_factor

    @property
    def inverse(self) -> Optional[functions.Function]:
        return ScalePositions(1. / self._scale_factor)

    @staticmethod
    def output_length(in_state: AtomsCollection) -> int:
        return len(in_state)

    def evaluate(self, atoms: AtomsCollection, *, get_jacobian=False):  # pylint: disable=arguments-differ
        out_atoms = AtomsCollection(atoms.num_atoms, dtype=atoms.dtype)
        out_atoms.positions[:] = self._scale_factor * atoms.positions
        out_atoms.numbers[:] = atoms.numbers

        if get_jacobian:
            jac = np.zeros((self.output_length(atoms), len(atoms)))
            final_pos_idx = 3 * atoms.num_atoms
            np.fill_diagonal(jac[0:final_pos_idx, 0:final_pos_idx], self._scale_factor)
            np.fill_diagonal(jac[final_pos_idx:, final_pos_idx:], 1.)
            return out_atoms, jac

        return out_atoms


class CentreAtomsCollection(functions.Function):
    """Centre an atomic system by translating it such that the centroid is coincident with the origin"""

    input_type = AtomsCollection
    output_type = AtomsCollection
    supports_jacobian = False

    @property
    def inverse(self) -> Optional[functions.Function]:
        """It is not possible to 'uncenter' a set of atoms (as we dont' keep track of where the previous centre was)
        and so the inverse of this function simply does nothing."""
        return functions.Identity()

    @staticmethod
    def output_length(in_state: AtomsCollection) -> int:
        return len(in_state)

    def evaluate(self, in_atoms: AtomsCollection, get_jacobian=False) -> AtomsCollection:  # pylint: disable=arguments-differ
        out_atoms = AtomsCollection(in_atoms.num_atoms)

        centre = np.sum(in_atoms.positions) / in_atoms.num_atoms
        out_atoms.positions = in_atoms.positions - centre
        out_atoms.numbers = in_atoms.numbers

        return out_atoms


class MapNumbers(functions.Function):
    """Map the given set of atom numbers onto a continuous range"""
    input_type = AtomsCollection
    output_type = AtomsCollection
    supports_jacobian = False

    def __init__(self, species: set, map_to: Union[float, Tuple[float, float]] = (1.0, 6.0)):
        super().__init__()
        if not isinstance(map_to, tuple):
            # Assume it's a scalar and everything is being mapped to a single number
            map_to = (map_to, map_to)

        species = set(species)
        self._numbers = list(sorted(species))
        self._mapped_range = map_to
        self._range_size = map_to[1] - map_to[0]
        self._half_bin = self._range_size / (2 * len(self._numbers))

    @property
    def numbers(self) -> List[int]:
        return self._numbers

    @property
    def mapped_range(self) -> Tuple[float, float]:
        """Get the range that species are mapped onto"""
        return self._mapped_range

    @property
    def inverse(self) -> Optional['functions.Function']:
        return MapNumbers.Inverse(set(self._numbers), self._mapped_range)

    @staticmethod
    def output_length(in_state: AtomsCollection) -> int:
        return len(in_state)

    def evaluate(self, in_atoms: AtomsCollection, get_jacobian=False) -> AtomsCollection:  # pylint: disable=arguments-differ
        out_atoms = in_atoms.copy()

        # Now adjust the numbers
        for idx, num in enumerate(out_atoms.numbers):
            try:
                num_idx = self._numbers.index(num)
            except ValueError:
                pass
            else:
                if num is not None:
                    new_numbers = (num_idx / len(self._numbers)) * self._range_size + self._mapped_range[0]
                    out_atoms.numbers[idx] = new_numbers + self._half_bin

        return out_atoms

    class Inverse(functions.Function):

        def __init__(self, possible_numbers: set, mapped_range: Tuple[float, float]):
            super().__init__()
            self._numbers = list(sorted(possible_numbers))
            self._mapped_range = mapped_range
            self._range_size = mapped_range[1] - mapped_range[0]

        @staticmethod
        def output_length(in_state: AtomsCollection) -> int:
            return len(in_state)

        def evaluate(self, in_atoms: AtomsCollection, get_jacobian=False) -> AtomsCollection:  # pylint: disable=arguments-differ
            out_atoms = in_atoms.copy()

            # Now adjust the numbers
            for idx, num in enumerate(out_atoms.numbers):
                if self._mapped_range[0] <= num <= self._mapped_range[1]:
                    rescaled = (num - self._mapped_range[0]) / self._range_size * len(self._numbers)
                    out_atoms.numbers[idx] = self._numbers[int(rescaled)]
                else:
                    _LOGGER.warning(
                        'Got a species number that is no in the range: %i <= %i <= %i', self._mapped_range[0], num,
                        self._mapped_range[1]
                    )

            return out_atoms


class ApplyCutoff(functions.Function):
    """Given a collection of atoms exclude any that are further from the origin than the given cutoff"""
    input_type = AtomsCollection

    def __init__(self, cutoff: float):
        super().__init__()
        self._cutoff_sq = cutoff * cutoff

    @property
    def inverse(self) -> Optional[functions.Function]:
        """Naturally, this function is not fully invertible as atoms that have been cut off cannot be replaced the
        inverse is just the identity"""
        return functions.Identity()

    def evaluate(self, in_atoms: AtomsCollection, get_jacobian=False):  # pylint: disable=arguments-differ
        index_map = {}
        # Find all those that are within the cutoff
        for idx in range(in_atoms.num_atoms):
            pos = in_atoms.positions[idx]
            if np.dot(pos, pos) < self._cutoff_sq:
                index_map[idx] = len(index_map)
            else:
                print('WARNING: CUTTING OFF {}'.format(idx))

        out_atoms = AtomsCollection(
            len(index_map),
            positions=in_atoms.positions[tuple(index_map.keys()), :],
            numbers=in_atoms.numbers[(tuple(index_map.keys()),)]
        )

        if get_jacobian:
            jac = np.zeros((len(out_atoms), len(in_atoms)))
            for old_idx, new_idx in index_map.items():
                jac[out_atoms.linear_pos_idx(new_idx), in_atoms.linear_pos_idx(old_idx)] = np.eye(3)
                jac[out_atoms.linear_number_idx(new_idx), in_atoms.linear_number_idx(old_idx)] = 1.

            return out_atoms, jac

        return out_atoms


def random_atom_collection_in_sphere(num: int, radius=1., centre=True, numbers=1.) -> AtomsCollection:
    pts = generate.random_points_in_sphere(num, radius=radius, centre=centre)
    atoms = AtomsCollection(num, positions=pts)
    if isinstance(numbers, tuple):
        atoms.numbers[:] = random.choices(numbers, k=num)
    else:
        atoms.numbers[:] = numbers

    return atoms


class SeparationForce(functions.Function):
    """A separation force between atoms that come within the cutoff.  This can be used as a penalty in a loss function
    to """

    def __init__(self, force_constant: float, cutoff: float):
        super().__init__()
        self._force_constant = force_constant
        self._cutoff = cutoff

    def evaluate(self, atoms: AtomsCollection, *, get_jacobian=False):  # pylint: disable=arguments-differ
        distances = pdist(atoms.positions)
        total_energy = np.sum(tuple(map(self.energy, distances)))

        if get_jacobian:
            n = atoms.num_atoms  # pylint: disable=invalid-name
            jacobian = np.zeros((1, len(atoms)))
            positions = atoms.positions
            # Indexing to extract the ij distance from the pdist vec, see:
            # https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.distance.pdist.html#scipy.spatial.distance.pdist
            idx = lambda i, j: n * i + j - ((i + 2) * (i + 1)) // 2
            for i in range(n):
                for j in range(i, n):
                    dr = positions[j] - positions[i]  # pylint: disable=invalid-name
                    dist = distances[idx(i, j)]
                    force = self.force(dist) * dr / dist

                    jacobian[0, atoms.linear_pos_idx(i)] += force
                    jacobian[0, atoms.linear_pos_idx(j)] -= force

            return total_energy, jacobian

        return total_energy

    def energy(  # pylint: disable=invalid-name
        self, r: float
    ) -> float:
        """
        Energy is
            k (r^2 / 2 - r r0) + r0
        where k is force constant and r0 is cutoff

        :param r: separation between atoms
        :return: the corresponding energy
        """
        if r > self._cutoff:
            return 0.

        return self._force_constant * r * (0.5 * r - self._cutoff) + self._cutoff

    def force(  # pylint: disable=invalid-name
        self, r: float
    ):
        """
        Force is
            -k (r - r0)
        where k is force constant and r0 is cutoff

        :param r: separation between atoms
        :return: the corresponding force magnitude
        """
        if r > self._cutoff:
            return 0.

        return -self._force_constant * (r - self._cutoff)



# -*- coding: utf-8 -*-
import numpy as np
import pytest
import sympy

from milad import analytic
from milad import functions
from milad import generate
from milad import geometric
from milad import utils

# pylint: disable=invalid-name


def test_multidim_norm_moments():
    sigma = 2.

    moms = geometric.gaussian_moment[0](np.zeros((3, 1)), sigma)
    expected = np.empty((3, 1))
    expected.fill(1.)
    assert pytest.approx((moms - expected).max(), 0)

    # Do the odds first
    for order in (1, 3, 5, 7):
        moms = geometric.gaussian_moment[order](np.zeros((3, 1)))
        expected = np.zeros((3, 1))
        assert pytest.approx((moms - expected).max(), 0)

    moms = geometric.gaussian_moment[2](np.zeros((3, 1)), sigma)
    expected = np.empty((3, 1))
    expected.fill(sigma**2)
    assert pytest.approx((moms - expected).max(), 0)

    moms = geometric.gaussian_moment[4](np.zeros((3, 1)), sigma)
    expected = np.empty((3, 1))
    expected.fill(3 * sigma**4)
    assert pytest.approx((moms - expected).max(), 0)

    moms = geometric.gaussian_moment[6](np.zeros((3, 1)), sigma)
    expected = np.empty((3, 1))
    expected.fill(15 * sigma**6)
    assert pytest.approx((moms - expected).max(), 0)

    moms = geometric.gaussian_moment[8](np.zeros((3, 1)), sigma)
    expected = np.empty((3, 1))
    expected.fill(105 * sigma**8)
    assert pytest.approx((moms - expected).max(), 0)


def test_moment_tensor3d():
    pos = np.array((1., 2., 3.))
    sigma = 2.
    mass = 1.
    max_order = 4

    moments = geometric.from_gaussians(max_order, [pos], sigma, mass)
    assert moments[0, 0, 0] == mass

    assert moments[1, 0, 0] == pos[0]
    assert moments[2, 0, 0] == pos[0]**2 + sigma**2
    assert moments[3, 0, 0] == pos[0]**3 + 3 * pos[0] * sigma**2

    assert moments[0, 1, 0] == pos[1]
    assert moments[0, 2, 0] == pos[1]**2 + sigma**2
    assert moments[0, 3, 0] == pos[1]**3 + 3 * pos[1] * sigma**2

    assert moments[0, 0, 1] == pos[2]
    assert moments[0, 0, 2] == pos[2]**2 + sigma**2
    assert moments[0, 0, 3] == pos[2]**3 + 3 * pos[2] * sigma**2

    for p, q, r in moments.iter_indices():
        assert moments[p, q, r] == moments[p, 0, 0] * moments[0, q, 0] * moments[0, 0, r]


def test_moments_symmetric():
    x = 2.
    positions = np.array(((-x, 0., 0.), (x, 0., 0.)))
    sigma = 2.
    mass = 1.5
    max_order = 4

    tensor = geometric.from_gaussians(max_order, positions, sigma, mass)
    assert tensor[0, 0, 0] == mass * len(positions)
    assert tensor[1, 0, 0] == 0.
    assert tensor[2, 0, 0] == 2 * mass * (x**2 + sigma**2)


def test_geom_moments_of_deltas():
    num_points = 4
    positions = generate.random_points_in_sphere(num_points, radius=.7)
    weights = np.random.rand(num_points)
    max_order = 11

    # Manually calculate the moments
    moms = np.zeros((max_order + 1, max_order + 1, max_order + 1))
    for p in utils.inclusive(max_order):
        for q in utils.inclusive(max_order):
            for r in utils.inclusive(max_order):
                for pos, weight in zip(positions, weights):
                    moms[p, q, r] += weight * (pos**(p, q, r)).prod(axis=-1)

    calculated = geometric.from_deltas(max_order, positions, weights=weights)
    np.testing.assert_array_almost_equal(moms, calculated.to_matrix())

    # The 0^th moment should always be the sum of the weights
    assert calculated[0, 0, 0] == weights.sum()


def test_delta_moments_derivatives():
    """Test that derivatives of moments calculated from delta functions are correct"""
    ORDER = 5
    NUM_POINTS = 4
    Delta = functions.WeightedDelta

    x = sympy.IndexedBase('x')
    w = sympy.IndexedBase('w')

    POINTS = analytic.create_array(x, (NUM_POINTS, 3))
    WEIGHTS = analytic.create_array(w, NUM_POINTS)

    # Get the moments and derivatives wrt to weights and positions
    moments, jac = geometric.from_deltas(ORDER, POINTS, weights=WEIGHTS, get_jacobian=True)

    for i in range(NUM_POINTS):
        delta_idx = i * Delta.LENGTH
        for j, (p, q, r) in enumerate(moments.iter_indices()):
            moment = moments[p, q, r]

            # Check weight derivatives
            dm_dx_calculated = jac[j, delta_idx + Delta.WEIGHT]
            if isinstance(moment, float):
                # We have a constant, so derivative is always 0
                dw_dx_analytic = 0
            else:
                dw_dx_analytic = moment.diff(w[i])

            assert dm_dx_calculated == dw_dx_analytic

            # Now loop over each x, y, z coordinate
            for dim in range(3):  # Look over x = 0, y = 1, z = 2
                dm_dx_calculated = jac[j, delta_idx + dim]
                if isinstance(moment, (float, sympy.core.numbers.Zero)):
                    # We have a constant, so derivative is always 0
                    dw_dx_analytic = 0
                else:
                    dw_dx_analytic = moment.diff(x[i, dim])

                assert dm_dx_calculated == dw_dx_analytic


def test_geometric_moments_calculator_deltas():
    """Test using the geometric moments calculator with delta functions as features"""
    num_points = 4
    positions = generate.random_points_in_sphere(num_points, radius=.7)
    weights = np.random.rand(num_points)
    max_order = 11

    # First, try a single delta function
    delta = functions.WeightedDelta(positions[0], weights[0])
    calculator = geometric.GeometricMomentsCalculator(max_order)
    moms = calculator(delta)
    moms2 = geometric.from_deltas(max_order, positions[0:1], weights[0:1])
    assert np.all(moms.moments == moms2.moments)

    # Now try all of them
    env = functions.Features(delta)
    for idx in range(1, num_points):
        env.add(functions.WeightedDelta(positions[idx], weights[idx]))

    moms = calculator(env)
    moms2 = geometric.from_deltas(max_order, positions, weights)
    assert np.all(moms.moments == moms2.moments)


def test_geometric_moments_calculator_gaussians():
    """Test using the geometric moments calculator with Gaussian functions as features"""
    num_points = 4
    positions = generate.random_points_in_sphere(num_points, radius=.7)
    sigmas = np.random.rand(num_points)
    weights = np.random.rand(num_points)
    max_order = 11

    # First, try a single delta function
    gaussian = functions.WeightedGaussian(positions[0], sigmas[0], weights[0])
    calculator = geometric.GeometricMomentsCalculator(max_order)
    moms = calculator(gaussian)
    moms2 = geometric.from_gaussians(max_order, positions[0:1], sigmas[0:1], weights[0:1])
    assert np.all(moms.moments == moms2.moments)

    # Now try all of them
    env = functions.Features(gaussian)
    for idx in range(1, num_points):
        env.add(functions.WeightedGaussian(positions[idx], sigmas[idx], weights[idx]))

    moms = calculator(env)
    moms2 = geometric.from_gaussians(max_order, positions, sigmas, weights)
    assert np.all(moms.moments == moms2.moments)



# -*- coding: utf-8 -*-
"""Module that is concerned with the calculation of moment invariants"""
import collections
import functools
import operator
import pathlib
from typing import Sequence, Union, List, Set, Tuple, Dict, Iterator, Callable

import numba
import numpy as np

from . import base_moments
from . import functions
from . import geometric

__all__ = ('MomentInvariant', 'read_invariants', 'RES_DIR', 'COMPLEX_INVARIANTS', 'GEOMETRIC_INVARIANTS', \
          'MomentInvariants')

# The resources directory
RES_DIR = pathlib.Path(__file__).parent / 'res'
GEOMETRIC_INVARIANTS = RES_DIR / 'rot3dinvs8mat.txt'
COMPLEX_INVARIANTS = RES_DIR / 'cmfs7indep_0.txt'
COMPLEX_INVARIANTS_ORIG = RES_DIR / 'cmfs7indep_0.orig.txt'
# A convenience map to make it easier to load the default invariants
INVS_MAP = {'geometric': GEOMETRIC_INVARIANTS, 'complex': COMPLEX_INVARIANTS, 'complex-orig': COMPLEX_INVARIANTS_ORIG}


def prod(iterable):
    return functools.reduce(operator.mul, iterable, 1)


class MomentInvariant:
    """Class storing moment invariants.

    The invariants consist of a sum of terms where each term has a prefactor multiplied by a product
    of moments labelled by three indices e.g. p * c_20^0 * c_20^0
    """

    def __init__(self, weight):
        self._weight = weight
        self._terms = []
        self._max_order = -1
        self._constant = 0

        self._farray = None  # The prefactors array
        self._indarray = None  # The index array
        self._norm_power = None
        self._derivatives = None  # A cache for the derivatives
        self._variables = None

    def __str__(self):
        sum_parts = []
        for prefactor, product in self._terms:
            powers = self._collect_powers(product)
            product_parts = [str(prefactor)]
            product_parts.extend(
                'm{},{},{}^{}'.format(indices[0], indices[1], indices[2], power) for indices, power in powers.items()
            )

            string = ' '.join(product_parts)
            sum_parts.append(string)

        return ' + '.join(sum_parts)

    @property
    def weight(self) -> int:
        """The number of terms in each product of the invariant.  This gives the units"""
        return self._weight

    @property
    def terms(self) -> List:
        return self._terms

    @property
    def max_order(self) -> int:
        if self._max_order == -1 and self._terms:
            term = self._terms[0]
            for indices in term[1]:
                self._max_order = max(self._max_order, np.max(indices))
        return self._max_order

    @property
    def variables(self) -> Set[Tuple]:
        """Get the indexes of all the unique moments used in this invariant e.g. if the invariant
        was:

            m_200 * m_200 * m_020 m_020 + m_002 m_002 + 2 m_110 m_110 + 2 m_101 m_101 + 2 m_011 m_011

        then this function would return

            {(2, 0, 0), (0, 2, 0), (0, 0, 2), (1, 1, 0), (1, 0, 0), (0, 1, 1)}

        as these are the all the moments (or variables) involved in calculating this invariant
        """
        if self._variables is None:
            variables = set()
            for _prefactor, product in self._terms:
                for part in product:
                    variables.add(part)
            self._variables = variables

        return self._variables

    @property
    def terms_array(self) -> np.ndarray:
        return self._indarray

    def insert(self, prefactor, indices: Sequence[Tuple]):
        """
        :param prefactor: the prefactor for this term in the invariant
        :param indices: the indices of the moments involved in this invariant
        """
        if not indices:
            # If there are no indices supplied then it's just a constant
            self._constant += prefactor
            return

        if not all(len(entry) == 3 for entry in indices):
            raise ValueError('There have to be three indices per entry, got: {}'.format(indices))
        self._terms.append((prefactor, tuple(indices)))

    def build(self):
        if self._terms:
            factors, arr = zip(*self._terms)
            self._farray = np.asarray(factors)
            self._indarray = np.asarray(arr)
            term = self._terms[0][1]
            self._norm_power = np.sum(term) / 3. + len(term)

    def apply(self, raw_moments: base_moments.Moments, normalise=False) -> float:
        """Compute this invariant from the given moments optionally normalising"""

        if isinstance(raw_moments, np.ndarray):
            total = self._numpy_apply(raw_moments)
        else:
            # If we can get a matrix we can still use the fast (numpy) method
            try:
                mtx = raw_moments.to_matrix()
            except AttributeError:
                # Ok, use generic method
                total = self._generic_apply(raw_moments)
            else:
                total = self._numpy_apply(mtx)

        if normalise:
            return total / raw_moments[0, 0, 0]**self._norm_power

        return total

    def __call__(self, *args, **kwargs):
        return self.apply(*args, **kwargs)

    # def _numpy_apply(self, raw_moments: np.ndarray):
    #     """Fast method to get the invariant from a numpy array"""
    #     total = self._constant  # type: float
    #
    #     if self._terms:
    #         indices = self._indarray
    #         total += np.prod(
    #             self._farray,
    #             np.product(raw_moments[indices[:, :, 0], indices[:, :, 1], indices[:, :, 2]],
    #                           axis=1)
    #         )
    #
    #     return total

    def _numpy_apply(self, raw_moments: np.ndarray):
        """Fast method to get the invariant from a numpy array"""
        total = self._constant  # type: float

        if self._terms:
            if raw_moments.dtype == object:
                # if raw_moments.dtype == object or len(self._terms) < 6:
                total += _numpy_apply(self._farray, self._indarray, raw_moments)
            else:
                total += _numba_apply(self._farray, self._indarray, raw_moments)

            #
            # if raw_moments.dtype == object or len(self._terms) < 48:
            #     # if raw_moments.dtype == object or len(self._terms) < 6:
            #     total += _numpy_apply(self._farray, self._indarray, raw_moments)
            # else:
            #     total += _parallel_apply(self._farray, self._indarray, raw_moments)

        return total

    def _generic_apply(self, moments):
        """Generic apply for moments that support indexing.

        This is slower version of above but compatible with moments that aren't numpy arrays"""
        total = self._constant
        for factor, indices in self._terms:
            product = 1
            for index in indices:
                product *= moments[index]
            total += factor * product
        return total

    def derivatives(self) -> Dict[Tuple, 'MomentInvariant']:
        """Get analytical derivatives for this invariant wrt each of its variables

        Returns a dictionary whose key is the variable (the index tuple) and the value the
        corresponding MomentInvariant
        """
        if not self._derivatives:
            # Have to calculate first time
            deriv_terms = {}  # type: Dict[Tuple, MomentInvariant]
            for prefactor, product in self._terms:
                powers = collections.defaultdict(int)
                for indices in product:
                    powers[indices] += 1

                # Now carry out the analytical derivative wrt each of our variables
                for variable in self.variables:
                    if variable not in powers:
                        continue

                    derivative = deriv_terms.setdefault(variable, MomentInvariant(self._weight - 1))
                    new_product = []

                    # Multiply the prefactor by the current power of the variable
                    power = powers[variable]
                    # Calculate the new prefactor
                    new_prefactor = prefactor * power

                    # And add in the correct multiple of this variable
                    if power != 1:
                        new_product.extend((variable,) * (power - 1))

                    for var, power in powers.items():
                        if var == variable:
                            # Skip this, we've dealt with it above
                            continue

                        # The other terms keep their exponent unchanged
                        new_product.extend((var,) * power)

                    derivative.insert(new_prefactor, new_product)

            for entry in deriv_terms.values():
                entry.build()

            self._derivatives = deriv_terms

        return self._derivatives

    @staticmethod
    def _collect_powers(product: List[Tuple]) -> Dict[Tuple, int]:
        powers = collections.defaultdict(int)
        for indices in product:
            powers[indices] += 1
        return powers


def _numpy_apply(prefactors, indices: np.array, raw_moments: np.ndarray):
    """Fast method to get the invariant from a numpy array"""
    total = 0.
    total += np.dot(prefactors, np.prod(raw_moments[indices[:, :, 0], indices[:, :, 1], indices[:, :, 2]], axis=1))

    return total


@numba.jit(parallel=False, nopython=True)
def _numba_apply(prefactors, indices, moments):
    """Generic apply for moments that support indexing.

    This is slower version of above but compatible with moments that aren't numpy arrays"""
    total = 0.
    for idx in numba.prange(len(prefactors)):  # pylint: disable=not-an-iterable
        factor = prefactors[idx]
        entry = indices[idx]

        product = 1.
        for index in entry:
            product *= moments[index[0], index[1], index[2]]

        total += factor * product
    return total


class MomentInvariants(functions.Function):
    """A function that takes moments as input and produces rotation invariants using polyonimals thereof"""
    input_type = base_moments.Moments
    output_type = np.ndarray
    supports_jacobian = True
    dtype = None

    def __init__(self, *invariant: MomentInvariant, are_real=True):
        super().__init__()
        for entry in invariant:
            if not isinstance(entry, MomentInvariant):
                raise TypeError(f'Expected MomentInvariant, got {entry.__class__.__name__}')

        self._invariants: List[MomentInvariant] = list(invariant)
        self._max_order = -1
        self._real = are_real

    def __len__(self) -> int:
        """Get the total number of invariants"""
        return len(self._invariants)

    def __iter__(self) -> Iterator[MomentInvariant]:
        """Iterate the invariants"""
        return self._invariants.__iter__()

    def __getitem__(self, item) -> Union['MomentInvariants', MomentInvariant]:
        """Get particular invariant(s)"""
        if isinstance(item, slice):
            return MomentInvariants(*self._invariants[item])
        if isinstance(item, tuple):
            if len(item) == 1:
                return MomentInvariants(self._invariants[item[0]])
            return MomentInvariants(*operator.itemgetter(*item)(self._invariants))

        return self._invariants[item]

    def filter(self, func: Callable) -> 'MomentInvariants':
        return MomentInvariants(*filter(func, self._invariants), are_real=self._real)

    def find(self, func: Callable) -> Tuple[int]:
        """Find the indices of invariants where fun(inv) returns True"""
        return tuple(i for i, inv in enumerate(self._invariants) if func(inv))

    @property
    def max_order(self) -> int:
        """Get the maximum order of all the invariants"""
        if self._max_order == -1:
            max_order = 0
            for inv in self._invariants:
                max_order = max(max_order, inv.max_order)
            self._max_order = max_order

        return self._max_order

    @property
    def variables(self) -> Set[Tuple]:
        """Return a set of all the the indices used by these invariants"""
        indices = set()
        for inv in self._invariants:
            indices.update(inv.variables)
        return indices

    def output_length(self, _in_state: functions.State) -> int:  # pylint: disable=unused-argument
        return len(self._invariants)

    def apply(self, moms: np.array, normalise=False, results=None) -> list:
        """Calculate the invariants from the given moments"""
        return apply_invariants(self._invariants, moms, normalise=normalise, results=results)

    def append(self, invariant: MomentInvariant):
        """Add an invariant"""
        self._invariants.append(invariant)
        self._max_order = max(self._max_order, invariant.max_order)

    def evaluate(self, moments: base_moments.Moments, *, get_jacobian=False) -> np.ndarray:  # pylint: disable=arguments-differ
        vector = np.empty(len(self._invariants), dtype=np.promote_types(moments.vector.dtype, float))
        jac = None
        if get_jacobian:
            jac = np.zeros((len(self._invariants), len(moments)), dtype=vector.dtype)

        for idx, inv in enumerate(self._invariants):
            vector[idx] = inv.apply(moments, normalise=False)

            if get_jacobian:
                # Evaluate the derivatives
                for index, dphi in inv.derivatives().items():
                    in_index = moments.linear_index(index)
                    jac[idx, in_index] = dphi.apply(moments)

        if self._real:
            vector = vector.real

        if get_jacobian:
            # Don't take 'real' of the Jacobian as if the moments are complex then the Jacobian
            # should be complex even though our output values are real
            return vector, jac

        return vector


def apply_invariants(invariants: List[MomentInvariant], moms: np.array, normalise=False, results=None) -> list:
    """Calculate the moment invariants for a given set of moments

    :param invariants: a list of invariants to calculate
    :param moms: the moments to use
    :param normalise: if True fill normalise the moments using the 0th moment
    :param results: an optional container to place the result in, if not supplied one will be created
    """
    if results is None:
        results = [None] * len(invariants)
    else:
        if not len(results) == len(invariants):
            raise ValueError('Results container must be of the same length as invariants')

    for idx, invariant in enumerate(invariants):
        results[idx] = invariant.apply(moms, normalise=normalise)

    return results


def read_invariants(filename: str = GEOMETRIC_INVARIANTS, read_max: int = None) -> \
        List[MomentInvariant]:
    """Read the Flusser, Suk and Zitová invariants.

    :param filename: the filename to read from, default to geometric moments invariants
    :param read_max: the maximum number of invariants to read
    """
    try:
        filename = INVS_MAP[filename]
    except KeyError:
        # Assume it is a filename
        pass

    invariants = []
    with open(filename, 'r') as file:

        for line in file:
            line = line.rstrip()
            if line:
                # New entry
                header = [int(number) for number in line.split(' ')]
                num_terms = header[2]
                invariant = MomentInvariant(num_terms)

                # Now read the actual terms
                line = file.readline().rstrip()
                while line:
                    terms = tuple(map(str_to_number, line.split(' ')))

                    prefactor = terms[0]

                    indices = []
                    for idx in range(num_terms):
                        indices.append(tuple(terms[idx * 3 + 1:(idx + 1) * 3 + 1]))
                    invariant.insert(prefactor, indices)

                    line = file.readline().rstrip()

                invariant.build()
                invariants.append(invariant)
                if len(invariants) == read_max:
                    break

    return invariants


def str_to_number(value: str) -> Union[int, float, complex]:
    """Convert an integer, float or complex number string to the correct number type"""
    try:
        return int(value)
    except ValueError:
        pass

    try:
        return float(value)
    except ValueError:
        pass

    try:
        return complex(value)
    except ValueError:
        pass

    raise ValueError('{} is not an int, float or complex'.format(value))


def read(filename: str = GEOMETRIC_INVARIANTS, read_max: int = None, max_order=None) -> \
        MomentInvariants:
    """Read the invariants from file"""
    invariants = MomentInvariants()
    try:
        filename = INVS_MAP[filename]
    except KeyError:
        # Assume it is a filename
        pass

    if filename == COMPLEX_INVARIANTS:
        invariants.dtype = complex

    for inv in read_invariants(filename, read_max):
        if max_order is None or inv.max_order <= max_order:
            invariants.append(inv)
    return invariants


def calc_moment_invariants(
    invariants: Sequence[MomentInvariant],
    positions: np.array,
    sigma: Union[float, np.array] = 0.4,
    masses: Union[float, np.array] = 1.,
    normalise=False
) -> Sequence[float]:
    """Calculate the moment invariants for a set of Gaussians at the given positions."""
    max_order = 0

    # Calculate the maximum order invariant we'll need
    for inv in invariants:
        max_order = max(max_order, inv.max_order)

    raw_moments = geometric.from_gaussians(max_order, positions, sigma, masses)
    return tuple(invariant.apply(raw_moments, normalise) for invariant in invariants)



# -*- coding: utf-8 -*-
import random

import numpy as np
import pytest

import milad
from milad import invariants

READ_MAX = 128  # The maximum number of invariants to read from the file


@pytest.fixture(scope='session')
def moment_invariants():
    """Get geometric moment invariants"""
    invs = milad.invariants.read(filename=invariants.GEOMETRIC_INVARIANTS, read_max=READ_MAX)
    yield invs


@pytest.fixture(scope='session')
def complex_invariants():
    """Get complex moment invariants"""
    invs = milad.invariants.read(filename=invariants.COMPLEX_INVARIANTS)
    yield invs


@pytest.fixture(autouse=True)
def set_random_seed():
    random.seed(1234)
    np.random.seed(1234)



# -*- coding: utf-8 -*-
from setuptools import setup

__author__ = 'Martin Uhrin'
__license__ = 'GPLv3'

about = {}
with open('milad/version.py') as f:
    exec(f.read(), about)  # pylint: disable=exec-used

setup(
    name='milad',
    version=about['__version__'],
    description='Moment Invariants Local Atomic Descriptor',
    long_description=open('README.rst').read(),
    url='https://github.com/muhrin/milad.git',
    author='Martin Uhrin',
    author_email='martin.uhrin.10@ucl.ac.uk',
    license=__license__,
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='machine learning, atomic descriptor, moment invariants',
    install_requires=[
        'dscribe', 'matplotlib', 'miniball', 'numba', 'numpy', 'rmsd', 'torch', 'scipy', 'sympy', 'scikit-learn',
        'scikit-image'
    ],
    extras_require={
        'dev': [
            'ipython',
            'pip',
            'pytest~=5.4',
            'pytest-cov',
            'pre-commit~=2.2',
            'prospector',
            'pylint==2.5.2',
            'twine',
            'yapf',
        ],
    },
    packages=['milad', 'milad.descriptors', 'milad.models', 'milad.optimisers', 'milad.play'],
    include_package_data=True,
    test_suite='test',
    entry_points={
        'milad.plugins.types': ['milad_types = milad.provides:get_mince_types'],
    }
)



# -*- coding: utf-8 -*-
import numpy as np


def to_real(vec: np.array) -> np.array:
    """Convert a given numpy vector containing complex numbers to one twice as long containing
    only real numbers where the first half contains the real and the second half the imaginary parts
    """
    view = vec.view('(2,)float')
    real = view.reshape(view.shape[0] * view.shape[1])
    return real


def to_complex(vec: np.array) -> np.array:
    """Given a vector of real numbers convert it to one half the size containing complex numbers
    where the first half of the original vector is treated as the real parts while the second half
    is used for the imaginary"""
    half_size = int(vec.size / 2)
    reshaped = vec.reshape((half_size, 2))
    view = reshaped.view(dtype=complex).reshape(half_size)

    return view


def even(val: int) -> bool:
    """Test if an integer is event.  Returns True if so."""
    return (val % 2) == 0



# -*- coding: utf-8 -*-
import math

import ase.build
import ase.io
import matplotlib.pyplot as plt
import numpy as np

import milad
from milad.play import asetools


def test_simple_atoms_fp(moment_invariants):
    cu_fcc = ase.build.bulk('Cu', 'sc', a=2.)
    ase.io.write('cu_sc.cif', cu_fcc)

    calculator = milad.play.FingerprintCalculator(moment_invariants, cutoff=2.5, cutoff_function=None)
    fps = asetools.calculate_fingerprint(cu_fcc, calculator)
    fps2 = asetools.calculate_fingerprints_dscribe(cu_fcc, calculator)

    diff = fps - fps2

    assert diff.max() == 0.
    assert len(fps) == 1


def test_multiple_species(moment_invariants, request):
    molecule = ase.build.molecule('CH3CH2OH')
    species = ['C', 'H', 'O']
    symbols = molecule.symbols
    num_atoms = len(symbols)

    fp = asetools.MiladFingerprint(species, moment_invariants, sigmas={'C': 1., 'H': 0.5, 'O': 0.7})

    fingerprints = fp.create(molecule)

    labels = []
    for i in range(num_atoms):
        labels.append('{}$_{}$'.format(i, symbols[i]))

    assert (fingerprints[3] - fp.create_single(molecule, 3)).max() == 0

    fig, axes = plt.subplots()
    axes.set_yscale('log')
    milad.plot.plot_multiple_invariants(fingerprints, axes, labels=labels)
    axes.set_title('C$_2$H$_6$O')
    fig.legend()
    fig.savefig('{}.pdf'.format(request.node.name))


def test_multiple_species_split(moment_invariants, request):
    molecule = ase.build.molecule('CH3CH2OH')
    species = ['C', 'H', 'O']

    fp = asetools.MiladFingerprint(
        species, moment_invariants, sigmas={
            'C': 0.4,
            'H': 0.3,
            'O': 0.5
        }, split_specie_pairs=True, cutoffs=100.
    )

    fingerprint = fp.create_single(molecule, 2)

    fig, axes = plt.subplots()
    axes.set_yscale('log')
    axes.set_title(str(molecule))
    milad.plot.plot_multiple_invariants(fingerprint, axes, labels=species)
    fig.legend()
    fig.savefig('{}.pdf'.format(request.node.name))


def test_generate_environments_molecule():
    molecule = ase.build.molecule('CH3CH2OH')

    envs = list(asetools.extract_environments(molecule, cutoff=4.))
    assert len(envs) == len(molecule)


def test_generate_environments_solid():
    lattice_param = 2.5
    cutoff = 4.

    diamond = ase.build.bulk('C', 'hcp', a=lattice_param)
    envs = list(asetools.extract_environments(diamond, cutoff=cutoff))

    lattice_max = int(math.ceil(lattice_param / cutoff) + 1)

    cutoff_sq = cutoff * cutoff
    # lattice params
    a, b, c = diamond.cell[0], diamond.cell[1], diamond.cell[2]

    found_positions = []
    for i, central in enumerate(diamond.positions):
        env_positions = []
        for j, other in enumerate(diamond.positions):
            r_ij = other - central

            for l in range(-lattice_max, lattice_max + 1):
                for m in range(-lattice_max, lattice_max + 1):
                    for n in range(-lattice_max, lattice_max + 1):
                        # if l == m == n == 0:
                        #     continue

                        dr = r_ij + l * a + m * b + n * c
                        if np.dot(dr, dr) < cutoff_sq:
                            env_positions.append(dr)

        known_positions = envs[i].positions
        assert len(env_positions) == len(known_positions)
        positions = np.array(env_positions)
        found_positions.append(positions)

        # Check the list of distances matches at least
        assert sorted(np.sum(positions * positions, axis=1)) == \
               sorted(np.sum(known_positions * known_positions, axis=1))

    assert len(envs) == len(found_positions)



# -*- coding: utf-8 -*-
from typing import Iterable

import numpy

import milad
from milad import reconstruct
from . import envs

__all__ = ('FingerprintCalculator',)


class FingerprintCalculator:

    def __init__(self, invariants=None, sigma=0.5, cutoff: float = 6., cutoff_function='cos', normalise=True):
        self._invariants = invariants or milad.invariants.read_invariants(read_max=10)
        self._sigma = sigma
        self._cutoff = cutoff
        self._cutoff_function = cutoff_function
        self._normalise = normalise

    @property
    def cutoff(self) -> float:
        return self._cutoff

    def calculate(self, positions: numpy.array) -> numpy.array:
        num = len(positions)
        invs_size = len(self._invariants)

        fingerprint = numpy.empty((num, invs_size))

        for i, pos in enumerate(positions):
            env = envs.SmoothGaussianEnvironment(pos, self._cutoff, self._cutoff_function)
            for other in positions:
                env.add_gaussian(other, self._sigma)

            fingerprint[i] = env.calc_moment_invariants(self._invariants, normalise=self._normalise)

        return numpy.reshape(fingerprint, (num * invs_size))

    def calculate_neighbours(self, pos: numpy.array, neighbours: Iterable[numpy.array]):
        env = envs.SmoothGaussianEnvironment(pos, self._cutoff, self._cutoff_function)
        for neighbour in neighbours:
            env.add_gaussian(neighbour, self._sigma)

        return env.calc_moment_invariants(self._invariants, normalise=self._normalise)



# -*- coding: utf-8 -*-
import copy
import functools

import numpy as np

from milad import base_moments
from milad import invariants as invs
from milad import utils
from . import least_squares
from . import root

__all__ = ('MomentsOptimiser',)


class MomentsOptimiser:
    """Given a vector of invariants this will attempt to recover the corresponding moments"""

    def __init__(self):
        self._least_squares_optimiser = least_squares.LeastSquaresOptimiser()
        # Use a less accurate tolerance while finding one by one and then crank up
        self._mask_invariant_moments = True
        self._root_finger = root.RootFinder()

    def optimise(  # pylint: disable=too-many-locals
            self,
            invariants_fn: invs.MomentInvariants,
            target: np.ndarray,
            initial: base_moments.Moments,
            jacobian='native',
            bounds=(-np.inf, np.inf),
            target_rmsd=1e-5,
            max_func_evals=5000,
            cost_tol=1e-5,
            grad_tol=1e-8,
            max_retries=2,
            verbose=False,
    ) -> least_squares.OptimiserResult:
        # Copy the start point
        current_moments = copy.deepcopy(initial)
        mask = initial.get_mask()
        mask.array.data[:, :, :] = current_moments.array[:, :, :]

        keep_fixed = {}  # Keep track of the ones we shouldn't unmask
        if self._mask_invariant_moments:
            keep_fixed = self._find_invariant_moments(invariants_fn, target)

        # Unmask 1st order
        mask.array.data[1, :, :] = None

        # Optimise from second order up
        for order in utils.inclusive(2, invariants_fn.max_order):
            indices = invariants_fn.find(functools.partial(max_n_degree, order))
            partial_invariants = invariants_fn[indices]
            partial_target = target[list(indices)]

            # Unmask the moments of this order
            mask.array.data[order, :, :] = None

            # Keep track of the best result for this order
            best_result = None
            for attempt in utils.inclusive(max_retries):
                if verbose:
                    print(f'Attempt {attempt}')

                # Let's mask off the degrees of freedom we know don't change
                self._fix_invariant_moments(mask, keep_fixed)

                result = self._least_squares_optimiser.optimise_target(
                    func=partial_invariants,
                    initial=current_moments,
                    target=partial_target,
                    mask=mask,
                    jacobian=jacobian,
                    bounds=bounds,
                    max_func_evals=128,
                    cost_tol=100 * cost_tol,
                    grad_tol=100 * grad_tol,
                    verbose=verbose
                )

                if verbose:
                    print(f'Found l={order} with RMSD: {result.rmsd}')

                # Keep track of the best result so far
                if best_result is None or result.rmsd < best_result.rmsd:
                    best_result = result

                if best_result.rmsd <= 100 * target_rmsd:
                    if verbose:
                        print('Keeping')
                    break

                # Going to have to try again: generate new moments of this order
                current_moments.randomise(indices=(order, None, None))

            # Keep the best results for the next order
            current_moments = best_result.value

        mask = current_moments.get_mask()
        self._fix_invariant_moments(mask, keep_fixed)

        # Now let's do one final optimisation to clean things up
        result = self._least_squares_optimiser.optimise_target(
            func=invariants_fn,
            initial=current_moments,
            target=target,
            mask=mask,
            jacobian=jacobian,
            bounds=bounds,
            max_func_evals=max_func_evals,
            cost_tol=cost_tol,
            grad_tol=grad_tol,
            verbose=verbose
        )

        if verbose:
            print(f'Found solution with RMSD: {result.rmsd}')

        return result

    def optimise4(  # pylint: disable=too-many-locals
            self,
            invariants_fn: invs.MomentInvariants,
            target: np.ndarray,
            initial: base_moments.Moments,
            jacobian='native',
            bounds=(-np.inf, np.inf),
            target_rmsd=1e-5,
            max_func_evals=5000,
            cost_tol=1e-5,
            grad_tol=1e-8,
            max_retries=1,
            verbose=False,
    ) -> least_squares.OptimiserResult:
        # Copy the start point
        current_moments = copy.deepcopy(initial)
        mask = initial.get_mask()
        mask.array.data[:, :, :] = current_moments.array[:, :, :]

        keep_fixed = {}  # Keep track of the ones we shouldn't unmask
        if self._mask_invariant_moments:
            keep_fixed = self._find_invariant_moments(invariants_fn, target)

        # Unmask first two orders
        mask.array.data[1:3, :, :] = None

        indices = invariants_fn.find(functools.partial(max_n_degree, 2))
        partial_invariants = invariants_fn[indices]
        partial_target = target[list(indices)]

        # Let's mask off the degrees of freedom we know don't change
        self._fix_invariant_moments(mask, keep_fixed)

        result = self._least_squares_optimiser.optimise_target(
            func=partial_invariants,
            initial=current_moments,
            target=partial_target,
            mask=mask,
            jacobian=jacobian,
            bounds=bounds,
            max_func_evals=128,
            cost_tol=100 * cost_tol,
            grad_tol=100 * grad_tol,
            verbose=verbose
        )
        # Copy over first and second order values
        mask.array.data[0:3, :, :] = result.value.array[0:3, :, :]

        done_invariants = set(indices)
        for order in utils.inclusive(3, invariants_fn.max_order):
            if verbose:
                print(f'==Doing up to N={order}')

            # Unmask this order
            mask.array.data[order, :, :] = None
            self._fix_invariant_moments(mask, keep_fixed)

            this_order = invariants_fn.find(functools.partial(max_n_degree, order))
            indices = tuple(set(this_order) - done_invariants)
            partial_invariants = invariants_fn[indices]
            partial_target = target[list(indices)]

            # Now use root finder to solve the rest
            result = root.RootFinder().optimise_target(
                partial_invariants, result.value, target=partial_target, mask=mask, max_func_evals=500, verbose=verbose
            )

            # Copy over found values
            mask.array.data[order, :, :] = result.value.array[order, :, :]
            done_invariants.update(set(indices))

        if verbose:
            print('==Doing final relaxation==')

        mask = current_moments.get_mask()
        self._fix_invariant_moments(mask, keep_fixed)

        # Now let's do one final optimisation to clean things up
        result = self._least_squares_optimiser.optimise_target(
            func=invariants_fn,
            initial=result.value,
            target=target,
            mask=mask,
            jacobian=jacobian,
            bounds=bounds,
            max_func_evals=max_func_evals,
            cost_tol=cost_tol,
            grad_tol=grad_tol,
            verbose=verbose
        )

        if verbose:
            print(f'Found solution with RMSD: {result.rmsd}')

        return result

    @staticmethod
    def _find_invariant_moments(invariants_fn: invs.MomentInvariants, invariants: np.ndarray) -> dict:
        # Let's mask off the degrees of freedom we know don't change
        invariant_moments = {}
        for inv_idx, inv in enumerate(invariants_fn):
            # If the invariant is equals just one moment then the moment itself is an invariant so mask it
            if len(inv.terms) == 1:
                idx = inv.terms[0][1][0]
                invariant_moments[idx] = invariants[inv_idx]

        return invariant_moments

    @staticmethod
    def _fix_invariant_moments(moments: base_moments.Moments, invariant_moments: dict):
        # Let's mask off the degrees of freedom we know don't change
        for mom_idx, inv_value in invariant_moments.items():
            if mom_idx[0] > moments.max_order:
                continue
            moments[mom_idx] = inv_value


def max_n_degree(max_n: int, invariant: invs.MomentInvariant):
    return invariant.terms_array[:, :, 0].max() <= max_n


def n_degree(n: int, invariant: invs.MomentInvariant):
    return np.all(invariant.terms_array[:, :, 0] == n)


def max_l_degree(max_l: int, invariant: invs.MomentInvariant):
    # The indexing is n, l, m so l is at index 1
    return invariant.terms_array[:, :, 1].max() <= max_l


def l_degree(  # pylint: disable=invalid-name
    l: int, invariant: invs.MomentInvariant
):
    # The indexing is n, l, m so l is at index 1
    return np.all(invariant.terms_array[:, :, 1] == l)


def containing_l_degree(  # pylint: disable=invalid-name
    l: int, invariant: invs.MomentInvariant
):
    return np.any(invariant.terms_array[:, :, 1] == l) and invariant.terms_array[:, :, 1].max() <= l



# -*- coding: utf-8 -*-
import random

import matplotlib.pyplot as plt
import numpy as np
import sympy

import milad
from milad import functions
from milad import generate
from milad import geometric


def test_invariant_single_mass(moment_invariants, request):
    num_invariants = 20
    num_masses = 20

    origin = np.zeros((1, 3))

    fig, axes = plt.subplots()

    for i in range(num_masses):
        mass = 0. + (0.1 * i)

        invariants = milad.invariants.calc_moment_invariants(moment_invariants[:num_invariants], origin, 0.4, mass)

        milad.plot.plot_invariants(invariants, axes, label='mass={}'.format(mass))

    fig.legend()
    fig.savefig('{}.pdf'.format(request.node.name))


def test_invariant_two_weights(moment_invariants, request):
    num_invariants = 64
    num_weights = 11

    positions = np.array(((-2., 0., 0.), (2., 0., 0.)))

    fig, axes = plt.subplots()

    for i in range(num_weights):
        mass = 0. + (0.1 * i)

        invariants = milad.invariants.calc_moment_invariants(
            moment_invariants[:num_invariants], positions, 0.4, (1., mass), normalise=True
        )

        milad.plot.plot_invariants(invariants, axes, label=f'$w={mass:.1f}$')

    axes.set_yscale('log')
    axes.set_title('Varying mass')
    fig.legend()
    fig.savefig('{}.pdf'.format(request.node.name))


def test_invariant_derivative(moment_invariants):
    # Take a random invariant and make sure that derivatives are getting calculated accurately
    invariant = random.choice(moment_invariants)

    # Symbols for moments
    m = sympy.IndexedBase('m')  # pylint: disable=invalid-name
    phi = invariant.apply(m)  # Analytic expression for moments
    derivatives = invariant.derivatives()

    for indices, entry in derivatives.items():
        dm = m[indices]  # pylint: disable=invalid-name
        dphi_dm_analytic = phi.diff(dm)
        dphi_dm_calculated = entry.apply(m)

        assert dphi_dm_calculated == dphi_dm_analytic


def test_invariants_function(moment_invariants):
    num_points = 10
    max_order = 10

    invariants_fn = milad.invariants.MomentInvariants(*moment_invariants)
    moments_fn = geometric.GeometricMomentsCalculator(max_order)

    pts = generate.random_points_in_sphere(num_points)
    env = functions.Features(*map(functions.WeightedDelta, pts))
    moments = moments_fn(env)
    phi = invariants_fn(moments)

    # The 1st moment is always the total mass
    assert phi[0] == num_points

    # Now try the same thing using chain
    combined_fn = functions.Chain(moments_fn, invariants_fn)
    phi2, _jacobian = combined_fn(env, jacobian=True)

    assert np.all(phi == phi2)


def test_invariants_derivatives_correctness(complex_invariants):
    complex_invariants = complex_invariants[:10]
    # The moments
    m = sympy.IndexedBase('m', complex=True)  # pylint: disable=invalid-name
    # Fill with symbols
    zernike_moms = milad.ZernikeMoments.from_indexed(m, complex_invariants.max_order, dtype=object)

    phi, jac = complex_invariants(zernike_moms, jacobian=True)

    for i, inv in enumerate(phi):
        for j, (mom_idx, mom) in enumerate(zernike_moms.iter(redundant=True)):
            # Perform symbolic derivative
            diff = sympy.diff(inv, mom).expand()

            # Get derivative from the Jacobian
            from_jac = jac[i, j]
            if not isinstance(from_jac, int):
                from_jac = from_jac.expand()

            # Compare
            difference = diff - from_jac
            if not diff == from_jac:
                # If they differ, check that it's by a meaningful amount
                coeffs = np.array(tuple(difference.as_coefficients_dict().values()))
                np.testing.assert_array_almost_equal(coeffs, 0., decimal=10)
